import smtplib
from email.mime.text import MIMEText

class EmailClient:
    def __init__(self, smtp_server, smtp_port, login, password):
        self.server = smtplib.SMTP(smtp_server, smtp_port)
        self.server.starttls()  # 使用TLS
        self.server.login(login, password)
        self.from_address = login

    def send_email(self, to_address, subject, body, from_address=None):
        if not from_address:
            from_address = self.from_address
        
        msg = MIMEText(body)
        msg['Subject'] = subject
        msg['From'] = from_address
        msg['To'] = to_address

        try:
            self.server.sendmail(from_address, to_address, msg.as_string())
            print("Email sent successfully!")
        except Exception as e:
            print(f"Failed to send email: {e}")

    def close(self):
        self.server.quit()

# client = EmailClient("smtp.gmail.com", 587, "YOUR_LOGIN@gmail.com", "YOUR_PASSWORD")
# client.send_email("receiver@example.com", "Hello", "This is a test email.")
# client.close()





import numpy as np
from .types import NO_ACTION
from .frame import DataIter
from typing import TypeVar, Tuple
from abc import ABC, ABCMeta, abstractmethod
from collections import deque
import pandas as pd

ActType = TypeVar("ActType")
ObsType = TypeVar("ObsType")

BUFFER_SIZE = 50


class MDFIter:
    def __init__(self, data, data_list, prices):
        self.data: pd.DataFrame = data
        self.data_list = data_list
        self.prices = prices
        self.returns = prices.pct_change().fillna(0)
        self.idx = 0
        self.trade_dates = list(self.data.index.levels[0])
        self.symbols = list(self.data.index.levels[1])
        self.current_date = None
        self.current_data = None

    def next(self):
        self.current_date = self.trade_dates[self.idx]
        self.current_data = self.data_list[self.idx]
        self.idx += 1
        return self.current_data

    def reset(self):
        self.current_data = None
        self.current_date = None
        self.idx = 0

    @property
    def spot_price(self):
        return self.prices.iloc[self.idx - 1]
    
    @property
    def spot_return(self):
        return self.returns.iloc[self.idx - 1]


class StockEnv:
    def __init__(self, data: pd.DataFrame, data_list: list[pd.DataFrame], prices: pd.DataFrame):
        self._iter = MDFIter(data, data_list, prices)
        self._account = Account(self._iter)
        self._initialized = False
        self._positions = np.repeat(np.nan, len(self._iter.symbols))
        self._reward = np.nan

    def step(self, 
             action: ActType
        ) -> Tuple[ObsType, float, bool, bool, dict]:
        assert self._initialized

        if action is not NO_ACTION:
            self._positions = action

        self._account.positions_series.loc[self.now] = self._positions
        self._account.strategy_returns.loc[self.now] = self._reward

        self._iter.next()
        spot_return = self._iter.spot_return
        strategy_return = np.sum(spot_return * self._positions)
        
        self._positions *= (1 + spot_return)
        self._positions /= np.sum(self._positions)

        next_state = self._iter.current_data
        self._reward = strategy_return
        truncated = False
        terminated = False
        info = None
        return next_state, self._reward, truncated, terminated, info
    
    def _update_positions(self):
        pass

    @property
    def current_date(self):
        return self._iter.current_date
    
    @property
    def now(self):
        return self._iter.current_date

    @property
    def current_data(self):
        return self._iter.current_data
    
    def reset(self) -> Tuple[ObsType, dict]:
        self._iter.reset()
        self._account.reset()
        self._positions = np.repeat(np.nan, len(self._iter.symbols)) 
        self._reward = np.nan

        self._iter.next()
        self._initialized = True
        return self._iter.current_data, None
    
    def spot(self, symbol=None, field=None) -> ObsType:
        # return self.data_iter.spot(symbol, field)
        pass


class Account:
    def __init__(self, iter: MDFIter):
        self._iter = iter
        self.positions_series = pd.DataFrame(np.nan, index=iter.trade_dates, columns=iter.symbols, dtype=np.float64)
        self.strategy_returns = pd.Series(np.nan, index=iter.trade_dates)
    
    def rebalance(self, target_positions):
        price = self._iter.spot_price()
        pre_positions = self.positions_series.loc[self._pre_trade_date]

        fees = np.nansum(np.abs(target_positions - self.positions))
        self.positions = target_positions
        self.costs = price
        return fees
    
    def reset(self):
        self.positions_series = pd.DataFrame(np.nan, index=self._iter.trade_dates, columns=self._iter.symbols, dtype=np.float64)
        self.strategy_returns = pd.Series(np.nan, index=self._iter.trade_dates) 
    
    def update(self):
        pass
        
    def calculate_capital_gains(self):
        price = self.data_iter.spot_price()
        assert np.sum(np.isnan(price)) == 0, "Detected nan value in price"
        if self.positions is None:
            return 0
        elif np.sum(self.positions) == 0:
            return 0
        else:
            capital_gains = np.nansum((price / self.costs - 1) * self.positions)
            self.positions = (price / self.costs) * self.positions
            self.positions /= np.nansum(self.positions)
            self.costs = price
            return capital_gains


import argparse

class alphagensClient:
    def __init__(self):
        self.parser = argparse.ArgumentParser(description="Alphagens command line tool")
        self.subparsers = self.parser.add_subparsers(dest="command")
        self.parser.register
        # Define the "clear" command
        clear_parser = self.subparsers.add_parser("clear", help="clear the outputs")
        # Define the "set" command
        set_parser = self.subparsers.add_parser("set", help="Set something")
        # ... add arguments for set ...
        # Define the "list" command
        list_parser = self.subparsers.add_parser("list", help="List items")
        # ... add arguments for list ...
        # Define the "create" command
        
        # create_parser = self.subparsers.add_parser("create", help="create the dependent files")
        # create_parser.add_argument("")

        create_parser = self.subparsers.add_parser("create", help="create latex environment")
        create_parser.add_argument("path", type=str, help="Specify the path where item should be create")
        create_parser.add_argument("--title", type=str, required=True, help="Specify the the title of pdf")
        create_parser.add_argument("--name", type=str, required=True, help="Specify the name of pdf")
        
    def run(self):
        args = self.parser.parse_args()
        pass



from abc import ABC, abstractmethod
import numpy as np
from scipy.optimize import minimize


class AbstractPortfolioOptimization(ABC):
    
    def __init__(self, n_assets):
        self.solution = None
        self.n_assets = n_assets
    
    @abstractmethod
    def objective_function(self, weights, mu, cov):
        raise NotImplementedError("should implement in the derived class")

    @abstractmethod
    def constraints(self):
        raise NotImplementedError("should implement in the derived class")
    
    @abstractmethod
    def bounds(self):
        return tuple((0, 1) for _ in range(self.n_assets))

    def _optimize(self, mu, cov):
        self.solution = minimize(
            lambda weights: self.objective_function(weights, mu, cov), 
            self.initial_guess(), 
            method='SLSQP', 
            bounds=self.bounds(), 
            constraints=self.constraints()
            )

    def get_optimal_weights(self):
        if self.solution is not None:
            return self.solution.x
        else:
            return "No solution found yet. Please run the optimize method first."

    def initial_guess(self):
        init = np.random.uniform(0, 1, size=self.n_assets)
        return init / np.sum(init)

    def __call__(self, mu, cov):
        self._optimize(mu, cov)
        return self.solution.x


class MinVarPortfolio(AbstractPortfolioOptimization):

    def __init__(self, n_assets):
        super().__init__(n_assets)
    
    def objective_function(self, weights, mu, cov):
        portfolio_mean = weights.T @ mu
        portfolio_variance = weights.T @ cov @ weights
        max_obj = - portfolio_variance
        return - max_obj

    def constraints(self):
        return ({'type': 'eq', 
                 'fun': lambda weights: np.sum(weights) - 1})
    
    def bounds(self):
        return tuple((0, 1) for _ in range(self.n_assets))
    

class MeanVarPortfolio(AbstractPortfolioOptimization):

    def __init__(self, n_assets, target_return, mean_return):
        super().__init__(n_assets)
        self.target_return = target_return
        self.mean_return = mean_return

    def objective_function(self, weights, mu, cov):
        portfolio_mean = weights.T @ mu
        portfolio_variance = weights.T @ cov @ weights
        max_obj = - portfolio_variance
        return - max_obj

    def constraints(self):
        return ({'type': 'eq', 
                 'fun': lambda weights: np.sum(weights) - 1},
                {'type': 'eq',
                 'fun': lambda weights: weights.T @ self.mean_return - self.target_return})
    
    def bounds(self):
        return tuple((0, 1) for _ in range(self.n_assets))
    

class QuadUtilityPortfolio(AbstractPortfolioOptimization):
    def __init__(self, n_assets, penalty):
        super().__init__(n_assets)
        self.penalty_coefficient = penalty

    def objective_function(self, weights, mu, cov):
        portfolio_mean = weights.T @ mu
        portfolio_variance = weights.T @ cov @ weights
        max_obj =  portfolio_mean - 0.5 * self.penalty_coefficient * portfolio_variance
        return - max_obj

    def constraints(self):
        return ({'type': 'eq', 
                 'fun': lambda weights: np.sum(weights) - 1})
    
    def bounds(self):
        return tuple((0, 1) for _ in range(self.n_assets))


class AnalyticMinVarPortfolio(AbstractPortfolioOptimization):

    def __init__(self, n_assets):
        super().__init__(n_assets)

    def objective_function(self, weights, mu, cov):
        pass

    def constraints(self):
        pass

    def bounds(self):
        pass

    def __call__(self, mu, cov):
        iota = np.repeat(1, self.n_assets)
        try:
            inv_cov = np.linalg.inv(cov)
            target_position = (inv_cov @ iota) / (iota.T @ inv_cov @ iota)

        except np.linalg.LinAlgError:
            print("uninvertiable covariance matrix detected! returns the naive weights")
            target_position = np.repeat(1/self.n_assets, self.n_assets)

        return target_position


class AnalyticMeanVarPortfolio(AbstractPortfolioOptimization):

    def __init__(self, n_assets, target_return):
        super().__init__(n_assets)
        self.target_return = target_return

    def objective_function(self, weights, mu, cov):
        pass

    def constraints(self):
        pass

    def bounds(self):
        pass

    def __call__(self, mu, cov):
        iota = np.repeat(1, self.n_assets)
        try:
            inv_cov = np.linalg.inv(cov)
            A = mu.T @ inv_cov @ mu
            B = iota.T @ inv_cov @ mu
            C = mu.T @ inv_cov @ iota
            D = iota.T @ inv_cov @ iota

            y = D / (A*D - B*C) * inv_cov @ mu - C / (A*D - B*C) * inv_cov @ iota
            z = A / (A*D - B*C) * inv_cov @ iota - B / (A*D - B*C) * inv_cov @ mu
            target_position = self.target_return * y + z
        except np.linalg.LinAlgError:
            print("uninvertiable covariance matrix detected!")
            target_position = np.repeat(1/len(self.n_assets), self.n_assets)

        return target_position
    

class AnalyticQuadUtilityPortfolio(AbstractPortfolioOptimization):
    def __init__(self, n_assets, penalty):
        super().__init__(n_assets)
        self.penalty = penalty

    def objective_function(self, weights, mu, cov):
        pass

    def constraints(self):
        pass

    def bounds(self):
        pass

    def __call__(self, mu, cov):
        alpha = self.penalty
        iota = np.repeat(1, self.n_assets)

        try:
            inv_cov = np.linalg.inv(cov)
            gamma = ( iota.T @ inv_cov @ mu) / (iota.T @ inv_cov @ iota)
            target_position = ( inv_cov @ iota) / (iota.T @ inv_cov @ iota) + (1/alpha) * inv_cov @ (mu - gamma * iota)

        except np.linalg.LinAlgError:
            print("uninvertiable covariance matrix detected!")
            target_position = np.repeat(1/len(self.n_assets), self.n_assets)
        
        return target_position


import tushare as ts
import pandas as pd
from joblib import Parallel, delayed, Memory
from ..calendars import DEFAULT_CALENDAR

ts.set_token("4701ef6fe54d814aad7ec2ec77a62c9152a9c3521bde0626c9717470")
pro = ts.pro_api()

MAX_JOBS = 64
DEFAULT_START_DATE = "20100101"
DEFAULT_END_DATE = "20230901"

memory = Memory(".cache")


@memory.cache
def ex_get_one_stock(ts_code, start_date, end_date):
    df: pd.DataFrame = ts.pro_bar(ts_code=ts_code, adj='qfq', start_date=start_date, end_date=end_date)
    df['trade_date'] = pd.DatetimeIndex(df['trade_date'])
    df.sort_values("trade_date", inplace=True, ignore_index=True)
    return df

@memory.cache
def get_one_stock(ts_code, start_date, end_date):
    df = pro.daily(ts_code=ts_code, start_date=start_date, end_date=end_date)
    df['trade_date'] = pd.DatetimeIndex(df['trade_date'])
    df.sort_values("trade_date", inplace=True, ignore_index=True)
    return df

@memory.cache
def get_one_stock_daily_basic(ts_code, start_date, end_date):
    df = pro.daily_basic(**{
        "ts_code": ts_code,
        "trade_date": "",
        "start_date": start_date,
        "end_date": end_date,
        "limit": "",
        "offset": ""
    }, fields=[
        "ts_code",
        "trade_date",
        "close",
        "turnover_rate",
        "turnover_rate_f",
        "volume_ratio",
        "pe",
        "pe_ttm",
        "pb",
        "ps",
        "ps_ttm",
        "dv_ratio",
        "dv_ttm",
        "total_share",
        "float_share",
        "free_share",
        "total_mv",
        "circ_mv"
    ])
    df['trade_date'] = pd.DatetimeIndex(df['trade_date'])
    df.sort_values("trade_date", inplace=True, ignore_index=True)
    return df


@memory.cache
def get_one_stock_moneyflow(ts_code, start_date, end_date):
    df = pro.moneyflow(ts_code=ts_code, start_date=start_date, end_date=end_date)
    df['trade_date'] = pd.DatetimeIndex(df['trade_date'])
    df.sort_values("trade_date", inplace=True, ignore_index=True)
    return df

@memory.cache
def get_one_stock_tech_factor(ts_code, start_date, end_date):
    df = pro.stk_factor(**{
        "ts_code": "000001.SZ",
        "start_date": "",
        "end_date": "",
        "trade_date": "",
        "limit": "",
        "offset": ""
    }, fields=[
        "ts_code",
        "trade_date",
        "close",
        "open",
        "high",
        "low",
        "pre_close",
        "change",
        "pct_change",
        "vol",
        "amount",
        "adj_factor",
        "open_hfq",
        "open_qfq",
        "close_hfq",
        "close_qfq",
        "high_hfq",
        "high_qfq",
        "low_hfq",
        "low_qfq",
        "pre_close_hfq",
        "pre_close_qfq",
        "macd_dif",
        "macd_dea",
        "macd",
        "kdj_k",
        "kdj_d",
        "kdj_j",
        "rsi_6",
        "rsi_12",
        "rsi_24",
        "boll_upper",
        "boll_mid",
        "boll_lower",
        "cci"
    ])
    df['trade_date'] = pd.DatetimeIndex(df['trade_date'])
    df.sort_values("trade_date", inplace=True, ignore_index=True)
    return df


class Stock:
    @staticmethod
    def tech_factor(symbols, start_date=DEFAULT_START_DATE, end_date=DEFAULT_END_DATE):
        def worker(symbol):
            return get_one_stock_tech_factor(symbol, start_date, end_date)
        results = Parallel(n_jobs=MAX_JOBS, prefer="threads")(delayed(worker)(symbol) for symbol in symbols)
        return results

    @staticmethod
    def moneyflow(symbols, start_date=DEFAULT_START_DATE, end_date=DEFAULT_END_DATE):
        def worker(symbol):
            return get_one_stock_moneyflow(symbol, start_date, end_date)

        works_to_do = [delayed(worker)(symbol) for symbol in symbols] 
        results = Parallel(n_jobs=MAX_JOBS, prefer="threads")(works_to_do)
        return results

    @staticmethod
    def daily_basic(symbols, start_date=DEFAULT_START_DATE, end_date=DEFAULT_END_DATE):
        def worker(symbol):
            return get_one_stock_daily_basic(symbol, start_date, end_date)
    
        results = Parallel(n_jobs=MAX_JOBS, prefer="threads")(delayed(worker)(symbol) for symbol in symbols)

        return results

    @staticmethod
    def ex_history(symbols, start_date=DEFAULT_START_DATE, end_date=DEFAULT_END_DATE):

        def worker(symbol):
            return ex_get_one_stock(symbol, start_date, end_date)
        
        works_to_do = [delayed(worker)(symbol) for symbol in symbols]

        results = Parallel(n_jobs=MAX_JOBS, prefer="threads")(works_to_do)

        return results

    @staticmethod
    def history(symbols, start_date=DEFAULT_START_DATE, end_date=DEFAULT_END_DATE):
        
        def worker(symbol):
            return get_one_stock(symbol, start_date, end_date)

        works_to_do = [delayed(worker)(symbol) for symbol in symbols] 
        
        results = Parallel(n_jobs=MAX_JOBS, prefer="threads")(works_to_do)

        return results


@memory.cache
def _get_one_index(ts_code, start_date, end_date):
    df = pro.index_daily(ts_code=ts_code, start_date=start_date, end_date=end_date)
    df['trade_date'] = pd.DatetimeIndex(df['trade_date'])
    df.sort_values("trade_date", inplace=True)
    df.set_index("trade_date", inplace=True)
    df = df.reindex(DEFAULT_CALENDAR.sessions_in_range(start_date, end_date))
    return df


def index_history(symbols, start_date, end_date):
    def worker(symbol):
        return _get_one_index(symbol, start_date, end_date)
    
    results = Parallel(n_jobs=MAX_JOBS, prefer="threads")(delayed(worker)(symbol) for symbol in symbols)
    return results


def get_rf_ret(start_date, end_date, termination=10):
    df = pro.yc_cb(**{
        "ts_code": "",
        "curve_type": 0,
        "trade_date": "",
        "start_date": start_date,
        "end_date": end_date,
        "curve_term": 10,
        "limit": "",
        "offset": ""
    }, fields=[
        "trade_date",
        "yield",
    ])
    df['trade_date'] = pd.DatetimeIndex(df['trade_date'])
    df.sort_values("trade_date", inplace=True)
    df.set_index("trade_date", inplace=True)
    df = df.reindex(DEFAULT_CALENDAR.sessions_in_range(start_date, end_date))
    return df


class Index:
    @staticmethod
    def components(ts_code, end_date):
        code_maps = {
            "000300.SH": "399300.SZ",
            "399300.SZ": "399300.SZ",
        }
        ts_code = code_maps[ts_code]
        df = pro.index_weight(index_code=ts_code, end_date=end_date)
        df = df.sort_values(["trade_date", "con_code"], ignore_index=True)
        date = df["trade_date"].iloc[-1]
        df = df[df["trade_date"]==date]
        return pd.Series(df["weight"].values / 100, index=df["con_code"]) 
    
    @staticmethod
    def history(symbols, start_date, end_date):
        def worker(symbol):
            return _get_one_index(symbol, start_date, end_date)
        
        results = Parallel(n_jobs=MAX_JOBS, prefer="threads")(delayed(worker)(symbol) for symbol in symbols)

        return results
    

class Bond:
    pass




import pandas as pd
import tqdm
import sys
import time


class Metric:
    def __init__(self):
        pass

    def after_session_end():
        pass

class SimulationEngine:
    """模拟市场
    """
    def __init__(self, trade_dates, data_portal: DataPortal):
        self._trade_dates = trade_dates
        self.data_portal = data_portal
        self.iter = iter(trade_dates)
        self.current_date = None
        self.spot_prices = None
    
    def on_session_start(self):
        self._next()

    def _next(self):
        self.current_date = next(self.iter)
        self.spot_prices = self.data_portal.prices.loc[self.current_date] # 索引改进, 目前会消耗200ms

class SimulatedBroker:
    def __init__(self, engine: SimulationEngine):
        self._engine = engine
        self.limit_up = self._engine.data_portal.get_trading_constraints("limit_up")

    def trading_constraints(self):
        pass

    def parser_order(self, account: "Account"):
        try:
            nums = account._order_lists.pop()
            prices = self._engine.spot_prices.loc[nums.index] # 可以加入冲击模型得到fill_price

            constraints = self.limit_up.loc[self._engine.current_date]
            constraints = constraints[constraints]

            sell_orders = nums[nums < 0].index
            buy_orders = nums[nums > 0].index
            buy_orders_masked = buy_orders.intersection(constraints)

            nums[buy_orders_masked] = 0

            filled_orders = pd.concat([nums, prices], axis=1)
            filled_orders.columns = ["nums", "fill_price"]
            account._filled_order_lists.append(filled_orders)
        except IndexError:
            pass

class Account:
    """统计账户信息
    """
    def __init__(self, engine):
        self._engine = engine
        self.positions = pd.Series(0, index=data_portal.universe, dtype=int)
        self.cash = 100000000
        self._dirty = True
        self._order_lists = []
        self._filled_order_lists = []

    @property
    def current_date(self):
        return self._engine.current_date

    @property
    def spot_prices(self):
        return self._engine.spot_prices

    def _update(self):
        self._dirty = True
        try:
            filled_orders = self._filled_order_lists.pop()
            assets = filled_orders.index
            nums = filled_orders["nums"]
            fill_price = filled_orders["fill_price"]
            total_cost = (nums * fill_price).sum() # 还可以加入成本, 印花税
            self.cash -= total_cost
            self.positions[assets] += nums
        except IndexError:
            pass
    
    def on_session_start(self):
        pass
        # self._previous_total_returns = self.returns

    def on_session_end(self):
        self._update()

    def order(self, amounts: pd.Series):
        self._order_lists.append(amounts)

    def order_target_pct_to(self, target_weights: pd.Series):
        assert target_weights.sum() <= 1.001
        assets = target_weights.index
        portfolio_value = self.portfolio_value
        spot_prices = self._engine.spot_prices.loc[assets]
        nums = (target_weights * portfolio_value / spot_prices).fillna(0).astype(int) - self.positions.loc[assets]
        self.order(nums)

    @property
    def portfolio_value(self):
        return self.cash + (self.positions * self.spot_prices).sum()
    
    @property
    def current_portfolio_weight(self):
        pass

class BaseStrategy:
    def __init__(self, engine, broker, account):
        self.broker: SimulatedBroker = broker
        self.engine: SimulationEngine = engine
        self.account: Account = account

    @property
    def current_date(self):
        return self.engine.current_date
    
    @property
    def spot_prices(self):
        return self.engine.spot_prices
    
    def order(self, positions: pd.Series):
        return self.account.order(positions)
    
    def before_trading_end(self):
        raise NotImplementedError

    def _before_trading_end(self):
        self.before_trading_end()
        self.broker.parser_order(self.account)

    def run_backtest(self):
        with tqdm.tqdm(total=len(Context.trade_dates), desc="") as pbar:
            while True:
                try:
                    self.engine.on_session_start()
                    self._before_trading_end()
                    self.account.on_session_end()
                except StopIteration:
                    break
                finally:
                    pbar.update(1)

class BuyAndHold(BaseStrategy):
    def __init__(self, engine, broker, account):
        super().__init__(engine, broker, account)
    
    def before_trading_end(self):
        if self.current_date == Context.trade_dates[0]:
            self.account.order_target_pct_to(pd.Series(1, index=["000001"]))

engine = SimulationEngine(Context.trade_dates, data_portal)
broker = SimulatedBroker(engine)
account = Account(engine)
algo = BuyAndHold(engine, broker, account)


import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import numpy as np


class GradientColor:
    def __init__(self, start_color, end_color, num_levels):
        # 获取预定义颜色的RGB值
        start_color_rgb = mcolors.to_rgb(start_color)
        end_color_rgb = mcolors.to_rgb(end_color)
        # 创建颜色映射
        self.cmap = mcolors.LinearSegmentedColormap.from_list(
            "mycmap", [start_color_rgb, end_color_rgb]
        )
        self.gradient = np.linspace(0, 1, num_levels)
        self.colors = self.cmap(self.gradient)
        self.hex_colors = [mcolors.rgb2hex(color) for color in self.colors]

    def get_hex_colors(self):
        return self.hex_colors

    def print_hex_colors(self):
        for hex_color in self.hex_colors:
            print(hex_color)

    def plot_colors(self):
        fig, ax = plt.subplots(1, 1, figsize=(5, 2),
                                dpi=80, facecolor='w', edgecolor='k')
        for sp in range(len(self.hex_colors)):
            ax.add_patch(
                plt.Rectangle((sp, 0), 1, 1, facecolor=self.hex_colors[sp])
            )
        ax.set_xlim([0, len(self.hex_colors)])
        ax.set_ylim([0, 1])
        ax.set_xticks([])
        ax.set_yticks([])
        plt.show()


class ColorGenerator:
    def __init__(self, n_groups):
        self.n_groups = n_groups
        cmap = plt.get_cmap('coolwarm')
        colors = [cmap(i) for i in np.linspace(0, 1, self.n_groups)]
        self.data = [mcolors.rgb2hex(color) for color in colors]

    def __call__(self):
        return self.data



BASE_COLORS = {'strategy': '#de3633',
            'benchmark': '#80b3f6',
            'excess': '#f4b63f'}




class CICC:
    """
    CICC的配色
    """
    colors = {'red': '#7f2e2a',
              'grey': '#f2f2f2'}

    def __getitem__(self, key):
        return self.colors[key]

    def __repr__(self):
        return self.colors.__repr__()


class LinearAlgebraDoneRight:
    """
    LinearAlgebraDoneRight的配色
    """
    colors = {'darkyellow': '#fdf5a8',
              'lightyellow': '#fefacc',
              'darkblue': '#b7c8e6',
              'lightblue': '#e1e8f5'}

    def __getitem__(self, key):
        return self.colors[key]

    def __repr__(self):
        return self.colors.__repr__()


import numpy as np
import pandas as pd
# import torchqtm.op.functional as F
from typing import Optional
import matplotlib.pyplot as plt


class RiskEvaluator(object):
    def __init__(self,
                 returns: pd.Series,
                 weights: pd.DataFrame = None,
                 freq=None,
                 rfr=0,
                 benchmark=None):
        self.returns = returns
        self.weights = weights
        self.freq = freq
        self.rfr = rfr
        self.benchmark = benchmark
        self.perf = {}

    @property
    def sharpe(self):
        return sharpe(self.returns, self.freq, self.rfr)

    @property
    def mdd(self):
        series, val, longest_time = drawdown(self.returns)
        return series, val, longest_time
    
    def _monthly_returns(self):
        pass

    def _yearly_returns(self):
        pass

    def _annualized_mean(self):
        pass

    def _annualized_volatility(self):
        pass


# def ic(factor_scores: pd.DataFrame, forward_returns: pd.DataFrame, method='pearson') -> pd.Series:
#     """
#     :return: id series
#     """
#     assert factor_scores.shape == forward_returns.shape
#     rlt = F.cs_corr(factor_scores, forward_returns, method=method)
#     rlt.fillna(0, inplace=True)
#     return rlt


def drawdown(returns: pd.Series) -> (pd.Series, float, int):
    """
    :returns : drawdown series, max drawdown, longest drawdown
    """
    net_curve = (1+returns).cumprod()
    cmax = net_curve.cummax()
    rlt = net_curve / cmax - 1
    longest_drawdown = cmax.value_counts().max() - 1
    # if isinstance(returns, pd.Series):
    #     rlt = pd.Series(rlt, index=returns.index)
    # elif isinstance(returns, np.ndarray):
    #     rlt = np.array(rlt)
    return rlt, rlt.min(), longest_drawdown


def turnover(weights: pd.DataFrame) -> pd.Series:
    """
    calculate the amount traded at each rebalance rawdata as fraction of portfolio size.
    """
    weights.fillna(0, inplace=True)
    forward_weights = weights.shift(1)
    forward_weights.fillna(0, inplace=True)
    rlt = (weights - forward_weights).abs().sum(axis=1)
    return rlt


def sharpe(returns, annulized_factor=252, rfr=0):
    if returns.isna().sum() > 0:
        raise ValueError('period_returns contains nan values.')

    mean = returns.mean()
    std = returns.std()
    _annf = annulized_factor
    return mean / std * _annf ** 0.5


def calculate_drawdowns(strategy_returns):
    net_series = np.cumprod(strategy_returns + 1)

    highwater_mark = 1
    max_drawdown = 0
    drawdown_start = False  
    max_drawdown_duration = 0
    current_duration = 0 
    
    for value in net_series:
        
        if value > highwater_mark:
            highwater_mark = value
            
            if drawdown_start:
                if current_duration > max_drawdown_duration:
                    max_drawdown_duration = current_duration

                current_duration = 0
                drawdown_start = False
        else:
            drawdown_start = True
            current_duration += 1
            
        drawdown = (highwater_mark - value) / highwater_mark
        if drawdown > max_drawdown:
            max_drawdown = drawdown
    
    if current_duration > max_drawdown_duration:
        max_drawdown_duration = current_duration
        
    return max_drawdown, max_drawdown_duration


def reg_stats(strategy_return, benchmark_return, rf_return=0, annulize_factor=252):
    y = strategy_return - rf_return
    X = benchmark_return.reshape(-1, 1) - rf_return
    from sklearn.linear_model import LinearRegression
    model = LinearRegression()
    model.fit(X, y)
    annulized_alpha = model.intercept_ * annulize_factor
    beta = model.coef_[0]
    residuals = y - model.predict(X)
    annulized_residual_std = np.std(residuals) * np.sqrt(annulize_factor)
    
    return annulized_alpha, beta, annulized_residual_std


import argparse

class AlphagensCLI:

    def __init__(self):
        self.parser = argparse.ArgumentParser(description="Alphagens command line tool")
        self.subparsers = self.parser.add_subparsers(dest="command")

        # Define the "set" command
        set_parser = self.subparsers.add_parser("set", help="Set something")
        # ... add arguments for set ...

        # Define the "list" command
        list_parser = self.subparsers.add_parser("list", help="List items")
        # ... add arguments for list ...

        # Define the "install" command
        install_parser = self.subparsers.add_parser("install", help="Install something")
        
        # Add --type and --name arguments for install command
        install_parser.add_argument("--type", 
                                    type=str, 
                                    required=True, 
                                    help="Specify the type of the item to install")
        install_parser.add_argument("--name", 
                                    type=str, 
                                    required=True, 
                                    help="Specify the name of the item to install")

    def run(self):
        args = self.parser.parse_args()

        # Handle the arguments
        if args.command == "set":
            self.handle_set(args)
        elif args.command == "list":
            self.handle_list(args)
        elif args.command == "install":
            self.handle_install(args)
        else:
            self.parser.print_help()

    def handle_set(self, args):
        # Your set command implementation
        pass

    def handle_list(self, args):
        # Your list command implementation
        pass

    def handle_install(self, args):
        item_type = args.type
        item_name = args.name
        print(f"Installing {item_type}: {item_name}...")
        # ... Rest of your install command implementation ...

def main():
    cli = AlphagensCLI()
    cli.run()

if __name__ == "__main__":
    main()



raise NotImplemented
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression

from typing import overload
# from .algos import rank_1d, rank_2d
from scipy.stats import norm
import talib

# Arithmetic Operators


def abs(x):
    """absolute value of x"""
    if isinstance(x, pd.DataFrame):
        return np.abs(x)
    elif isinstance(x, np.ndarray):
        return np.abs(x)


def ceiling(X):
    """ceiling of x"""
    return np.ceil(X)


def floor(X):
    return np.floor(X)


def divide(X, Y):
    return np.divide(X, Y)


# TODO: Revise it

def add(X, Y):
    return np.add(X, Y)


def sub(X, Y):
    return np.subtract(X, Y)


def mul(X, Y):
    np.multiply(X, Y)


def pow(X, p):
    return np.power(X, p)


def less(X, Y):
    return np.less(X, Y)


def more(X, Y):
    return np.greater(X, Y)


def leq(X, Y):
    return np.less_equal(X, Y)


def geq(X, Y):
    return np.greater_equal(X, Y)


def eq(X, Y):
    return np.equal(X, Y)


# def exp(X):
#     if isinstance(X, np.ndarray):
#         return np.exp(X)
#     elif isinstance(X, pd.DataFrame):
#         return X.apply(np.exp, axis=1)


def inverse(X):
    return divide(1, X)


def log(X):
    return np.log(X)


def log_diff(X):
    return np.log(X / ts_delay(X, 1))


def nan_out(X, lower=-0.1, upper=0.1):
    """set returns outside of [lower, upper] to nan"""
    rlt = np.where(((X < lower) + (X > upper)) > 0, np.nan, X)
    if isinstance(rlt, np.ndarray):
        return rlt
    elif isinstance(rlt, pd.DataFrame):
        return pd.DataFrame(rlt, index=X.index, columns=X.columns)


def purify(X):
    rlt = np.where(np.isinf(X), np.nan, X)
    if isinstance(X, np.ndarray):
        return rlt
    elif isinstance(X, pd.DataFrame):
        return pd.DataFrame(rlt, index=X.index, columns=X.columns)


def sign(X):
    rlt = np.where(X > 0, 1, X)
    rlt = np.where(X < 0, -1, rlt)
    rlt = np.where(X == 0, 0, rlt)
    if isinstance(rlt, np.ndarray):
        return rlt
    elif isinstance(rlt, pd.DataFrame):
        return pd.DataFrame(rlt, index=X.index, columns=X.columns)


def sqrt(x):
    return np.sqrt(x)


def to_nan(X, value=0, reverse=False):
    if not reverse:
        rlt = np.where(X == value, np.nan, X)
    else:
        rlt = np.where(X != value, np.nan, X)
    if isinstance(rlt, np.ndarray):
        return rlt
    elif isinstance(rlt, pd.DataFrame):
        return pd.DataFrame(rlt, index=X.index, columns=X.columns)


def densify(X):
    # TODO: implement this
    pass


# Logical Operators
# TODO: 1231
def if_else(condition, value_if_true, value_if_false):
    rlt = np.where(condition, value_if_true, value_if_false)
    if isinstance(condition, np.ndarray):
        return rlt
    elif isinstance(condition, pd.DataFrame):
        return pd.DataFrame(rlt, index=condition.index, columns=condition.columns)


def logical_and(x, y):
    return np.logical_and(x, y)


def logical_or(x, y):
    return np.logical_or(x, y)


def negate(x):
    return np.negative(x)


def is_nan(X):
    return np.isnan(X)


def if_finite(X):
    rlt = np.where(np.isfinite(X), 1, 0)
    if isinstance(rlt, np.ndarray):
        return rlt
    elif isinstance(rlt, pd.DataFrame):
        return pd.DataFrame(rlt, index=X.index, columns=X.columns)


# Time Series Operators
from torchqtm.core.window.rolling import (
    roll_apply,
    roll_apply_max,
    roll_apply_min,
    roll_apply_mean,
    roll_apply_sum,
    roll_apply_rank,
)


def ts_apply(x, d, func):
    if isinstance(x, np.ndarray):
        return roll_apply(x, d, func)
    elif isinstance(x, pd.Series):
        return pd.Series(roll_apply(x, d, func), index=x.index, name=x.name)
    elif isinstance(x, pd.DataFrame):
        return pd.DataFrame(roll_apply(x.values, d, func), index=x.index, columns=x.columns)


def ts_mean(x, d):
    def aux_func(array, window_size):
        return roll_apply_mean(array, window_size)

    if isinstance(x, np.ndarray):
        return aux_func(x, d)
    elif isinstance(x, pd.Series):
        return pd.Series(aux_func(x.values, d), index=x.index, name=x.name)
    elif isinstance(x, pd.DataFrame):
        return pd.DataFrame(aux_func(x.values, d), index=x.index, columns=x.columns)


def ts_max(x, d):
    def aux_fun(array, window_size):
        return roll_apply_max(array, window_size)

    if isinstance(x, np.ndarray):
        return aux_fun(x, d)
    elif isinstance(x, pd.Series):
        return pd.Series(aux_fun(x.values, d), index=x.index, name=x.name)
    elif isinstance(x, pd.DataFrame):
        return pd.DataFrame(aux_fun(x.values, d), index=x.index, columns=x.columns)


def ts_min(x, d):
    def aux_func(array, window_size):
        return roll_apply_min(array, window_size)

    if isinstance(x, np.ndarray):
        return aux_func(x, d)
    elif isinstance(x, pd.Series):
        return pd.Series(aux_func(x.values, d), index=x.index, name=x.name)
    elif isinstance(x, pd.DataFrame):
        return pd.DataFrame(aux_func(x.values, d), index=x.index, columns=x.columns)


def ts_sum(x, d, mode='auto'):
    def aux_func(array, window_size):
        return roll_apply_sum(array, window_size, mode=mode)

    if isinstance(x, np.ndarray):
        return aux_func(x, d)
    elif isinstance(x, pd.Series):
        return pd.Series(aux_func(x.values, d), index=x.index, name=x.name)
    elif isinstance(x, pd.DataFrame):
        return pd.DataFrame(aux_func(x.values, d), index=x.index, columns=x.columns)


def ts_delay(x, d):
    if isinstance(x, pd.DataFrame) or isinstance(x, pd.Series):
        return x.shift(d)
    elif isinstance(x, np.ndarray):
        return pd.DataFrame(x).shift(d).values
    else:
        raise TypeError("Input should be a pandas DataFrame, Series or a numpy ndarray.")


def ts_delta(x, d):
    return sub(x, ts_delay(x, d))


def days_from_last_change(x, d):
    pass


def ts_weighted_delay(x, k=0.5):
    return add(k * x, (1-k) * ts_delay(x, 1))


# 这俩函数有点抽象, 后边想想怎么实现
# def hump(x, hump=0.01):
#     """
#     This operator limits amount and magnitude of changes in input
#     (thus reducing turnover). If input changed by less than a threshold
#      compared to previous days’ output, current output is the same as on the previous day.
#     """
#
#
#
# def hump_decay(x, p=0):
#     pass


def inst_tvr(x, d):
    """
    Total trading value / Total holding value in the past d days

    """
    pass


def jump_decay(x, d, sensitivity=0.5, force=0.1):
    """
    If there is a huge jump in current data compare to previous one, apply force:
    我认为作用是滤平变动
    """
    cond = abs(x-ts_delay(x, 1) > sensitivity * ts_std_dev(x, d))
    value_if_true = ts_delay(x, 1) + ts_delta(x, 1) * force
    value_if_false = x
    return if_else(cond, value_if_true, value_if_false)


def kth_element(x, d, k):
    pass


def last_diff_value(x, d):
    pass


def ts_arg_min(x, d):
    def aux_func(t):
        return np.nanargmin(t[::-1], axis=0)

    return ts_apply(x, d, aux_func)


def ts_arg_max(x, d):
    def aux_func(t):
        return np.nanargmax(t[::-1], axis=0)

    return ts_apply(x, d, aux_func)


def ts_av_diff(x, d):
    return sub(x, ts_mean(x, d))


def ts_backfill(x, d, k=1, ignore="NAN"):
    pass


def ts_co_kurtosis(y, x, d):
    pass


def _ts_corr_single(x, y, d):
    pass


def ts_corr(x, y, d):
    """
    Pearson correlation of x, y in the past d days.
    """
    if len(x.shape) == 2:
        rlt = pd.DataFrame(x).rolling(d).corr(pd.DataFrame(y))
    else:
        rlt = pd.Series(x).rolling(d).corr(pd.Series(y))
    if isinstance(x, np.ndarray):
        return rlt.values
    elif isinstance(x, pd.Series):
        return pd.Series(rlt.values, index=x.index, name=x.name)
    elif isinstance(x, pd.DataFrame):
        return pd.DataFrame(rlt.values, index=x.index, columns=x.columns)


def ts_co_skewness(y, x, d):
    pass


def ts_count_nans(x, d):
    def aux_func(t):
        return np.nansum(np.isnan(t), axis=0)

    return ts_apply(x, d, aux_func)


def ts_covariance(y, x, d):
    pass


def ts_decay_exp_window(x, d, factor):
    pass


def ts_decay_linear(x, d, dense=False):
    # TODO: 想想怎么改进这个矩阵乘法, 目前会产生大量的NA
    def aux_func(data_slice):
        filters = np.arange(1, d+1) / ( d * (d + 2) / 2 )
        return np.matmul(data_slice.T, filters)
    return ts_apply(x, d, aux_func)


def ts_std_dev(x, d):
    def aux_func(t):
        return np.nanstd(t, axis=0)

    return ts_apply(x, d, aux_func)


def ts_ir(x, d):
    return divide(ts_mean(x, d), ts_std_dev(x, d))


def ts_kurtosis(x, d):
    pass


def ts_max_diff(x, d):
    return sub(x, ts_max(x, d))


def ts_median(x, d):
    def aux_func(t):
        return np.nanmedian(x, axis=0)
    return ts_apply(x, d, aux_func)


def ts_min_diff(x, d):
    return x - ts_min(x, d)


def ts_min_max_cps(x, d, f=2):
    return sub(add(ts_max(x, d), ts_min(x, d)), f * x)


def ts_min_max_diff(x, d, f=0.5):
    return sub(x, f * (ts_max(x, d) + ts_min(x, d)))


def ts_moment(x, d, k=0):
    def aux_func(t):
        return np.nanmean(np.power(t, k), axis=0)

    return ts_apply(x, d, aux_func)


def ts_partial_corr(x, y, z, d):
    pass


def ts_percentage(x, d, percentage=0.5):
    pass


def ts_poly_regression(y, x, d, k=1):
    pass


def ts_product(x, d):
    def aux_func(t):
        return np.nanprod(x, axis=0)

    return ts_apply(x, d, aux_func)


def ts_rank(x, d, mode="auto"):
    def aux_func(array, window_size):
        return roll_apply_rank(array, window_size, method=mode)

    if isinstance(x, np.ndarray):
        return aux_func(x, d)
    elif isinstance(x, pd.Series):
        return pd.Series(aux_func(x.values, d), index=x.index, name=x.name)
    elif isinstance(x, pd.DataFrame):
        return pd.DataFrame(aux_func(x.values, d), index=x.index, columns=x.columns)


def ts_regression(y, x, d, lag=0, rettype=0):
    pass


def ts_returns(x, d):
    return (x - ts_delay(x, d)) / ts_delay(x, d)


def ts_scale(x, d, constant=0):
    return constant + divide(x - ts_min(x, d), ts_max(x, d) - ts_min(x, d))


def ts_skewness(x, d):
    pass


# ts_step(1), step(1)


def ts_theilsen(x, y, d):
    pass


def ts_triple_corr(x, y, z, d):
    pass


def ts_zscore(x, d):
    return divide(x - ts_mean(x, d), ts_std_dev(x, d))


def ts_entropy(x, d):
    pass


def ts_vector_neut(x, y, d):
    pass


def ts_vector_proj(x, y, d):
    pass


# ts_rank_gmean_amean_diff(input1, input2, input3,...,d)


def _ts_quantile_uniform(x, d):
    def aux_func(array, window_size):
        return ts_rank(array, window_size) / window_size

    if isinstance(x, np.ndarray):
        return aux_func(x, d)
    elif isinstance(x, pd.Series):
        return pd.Series(aux_func(x.values, d), index=x.index, name=x.name)
    elif isinstance(x, pd.DataFrame):
        return pd.DataFrame(aux_func(x.values, d), index=x.index, columns=x.columns)


def _ts_quantile_gaussian(x, d):
    def aux_fun(array, window_size):
        return norm.ppf(ts_rank(array, window_size) / window_size)

    if isinstance(x, np.ndarray):
        return aux_fun(x, d)
    elif isinstance(x, pd.Series):
        return pd.Series(aux_fun(x.values, d), index=x.index, name=x.name)
    elif isinstance(x, pd.DataFrame):
        return pd.DataFrame(aux_fun(x.values, d), index=x.index, columns=x.columns)


def ts_quantile(x, d, driver="uniform"):
    if driver == "uniform":
        return _ts_quantile_uniform(x, d)
    elif driver == "gaussian":
        return _ts_quantile_gaussian(x, d)
    else:
        return ValueError


# CrossSectionalOperators
def _cs_corr(X, Y):
    mean_x = np.nanmean(X, axis=1, keepdims=True)
    std_x = np.nanstd(X, axis=1, keepdims=True)
    mean_y = np.nanmean(Y, axis=1, keepdims=True)
    std_y = np.nanstd(Y, axis=1, keepdims=True)
    return np.nanmean((X - mean_x) / std_x * (Y - mean_y) / std_y, axis=1)


def cs_corr(X, Y, method="pearson"):
    if method == "pearson":
        rlt = _cs_corr(X, Y)
    elif method == "spearman":
        # temp_X = pd.DataFrame(X).rank(axis=1)
        # temp_Y = pd.DataFrame(Y).rank(axis=1)
        temp_X = rank_2d(X, axis=1)
        temp_Y = rank_2d(Y, axis=1)
        temp_X = temp_X / np.nanmax(temp_X, axis=1, keepdims=True)
        temp_Y = temp_Y / np.nanmax(temp_Y, axis=1, keepdims=True)
        rlt = _cs_corr(temp_X, temp_Y)
    else:
        raise ValueError("the method should be pearson or spearman")
    if isinstance(X, np.ndarray):
        return rlt
    elif isinstance(X, pd.DataFrame):
        return pd.Series(rlt, index=X.index, name=f'{method}')


def cs_rank(X):
    assert len(X.shape) == 2

    def aux_func(t):
        return rank_2d(t, axis=1)

    if isinstance(X, np.ndarray):
        return aux_func(X)
    elif isinstance(X, pd.DataFrame):
        return pd.DataFrame(aux_func(X.values), index=X.index, columns=X.columns)


def _regression_neut(Y, X):
    base_shape = Y.shape
    n_axis = len(Y.shape)
    data = np.concatenate([Y.reshape(*base_shape, 1), X.reshape(*base_shape, 1)], axis=n_axis)

    def aux_func(data_slice):
        y = data_slice[..., 0]
        x = data_slice[..., 1]
        y_demean = y - np.nanmean(y)
        x_demean = x - np.nanmean(x)
        residuals = y_demean - (np.nanmean(y_demean * x_demean) / np.nanvar(x_demean)) * x_demean
        return residuals

    return ts_apply(data, 1, aux_func)


def _regression_neuts(Y, others):
    pass
    # result = np.empty_like(Y)
    # for i in range(len(Y)):
    #     y = Y[i]
    #     X = np.concatenate([x[i].reshape(-1, 1) for x in others], axis=1)
    #     model = LinearRegression()
    #     model.fit(X, y)
    #     y_pred = model.predict(X)
    #     residuals = y - y_pred
    #     result[i] = residuals
    # return pd.DataFrame(np.array(result), index=Y.index, columns=Y.columns)


def regression_neut(Y, others):
    if not isinstance(others, list):
        if isinstance(Y, np.ndarray):
            return _regression_neut(Y, others)
        elif isinstance(Y, pd.Series):
            return pd.Series(_regression_neut(Y.values, others.values), index=Y.index, name=Y.name)
        elif isinstance(Y, pd.DataFrame):
            return pd.DataFrame(_regression_neut(Y.values, others.values), index=Y.index, columns=Y.columns)
    else:
        raise ValueError


def regression_proj(Y, others):
    return Y - regression_neut(Y, others)


def _winsorize(X, method, param):
    assert isinstance(X, np.ndarray)
    n_stocks = X.shape[-1]
    n_axis = len(X.shape)
    if method == 'quantile':
        lower_bound = np.nanquantile(X, param, axis=n_axis-1, keepdims=True)
        upper_bound = np.nanquantile(X, 1-param, axis=n_axis-1, keepdims=True)
    elif method == 'std':
        mean = np.nanmean(X, axis=n_axis-1, keepdims=True)
        std = np.nanstd(X, axis=n_axis-1, keepdims=True)
        lower_bound = mean - std * param
        upper_bound = mean + std * param
    else:
        raise ValueError('Unknown method')
    return np.clip(a=X,
                   a_min=np.repeat(lower_bound, n_stocks, axis=n_axis-1),
                   a_max=np.repeat(upper_bound, n_stocks, axis=n_axis-1))


def winsorize(X, method='quantile', param=0.05):
    aux_func = _winsorize
    if isinstance(X, np.ndarray):
        return aux_func(X, method, param)
    elif isinstance(X, pd.Series):
        return pd.Series(aux_func(X.values, method, param), index=X.index, name=X.name)
    elif isinstance(X, pd.DataFrame):
        return pd.DataFrame(aux_func(X.values, method, param), index=X.index, columns=X.columns)


def _normalize(X, useStd):
    assert isinstance(X, np.ndarray)
    n_axis = len(X.shape)
    n_stocks = X.shape[-1]
    x_mean = np.nanmean(X, axis=n_axis-1, keepdims=True)
    s_std = np.nanstd(X, axis=n_axis-1, keepdims=True)
    if useStd:
        return (X - np.repeat(x_mean, n_stocks, axis=n_axis-1)) / np.repeat(s_std, n_stocks, axis=n_axis-1)
    else:
        return X - np.repeat(x_mean, n_stocks, axis=n_axis-1)


def normalize(X, useStd=True):
    aux_func = _normalize
    if isinstance(X, np.ndarray):
        return aux_func(X, useStd)
    elif isinstance(X, pd.Series):
        return pd.Series(aux_func(X.values, useStd), index=X.index, name=X.name)
    elif isinstance(X, pd.DataFrame):
        return pd.DataFrame(aux_func(X.values, useStd), index=X.index, columns=X.columns)


# # Group Operators
# ---
# Should be optimized
def _group(X, groups, agg_func):
    assert X.shape == groups.shape
    assert len(X.shape) == 2
    result = np.empty_like(X)
    if isinstance(X, pd.DataFrame):
        for i in range(len(X)):
            x, group = X.iloc[i], groups.iloc[i]
            result[i] = (x - x.groupby(group).transform(agg_func))
        return pd.DataFrame(result, index=X.index, columns=X.columns)
    elif isinstance(X, np.ndarray):
        X = pd.DataFrame(X)
        groups = pd.DataFrame(groups)
        for i in range(len(X)):
            x, group = X.iloc[i], groups.iloc[i]
            result[i] = (x - x.groupby(group).transform(agg_func))
        return result


def group_backfill(x, grou, d, std=4):
    pass


def group_count(x, group):
    pass


def group_extra(x, weight, group):
    pass


def group_mean(x, group):
    return _group(x, group, np.mean)


def group_max(x, group):
    return _group(x, group, np.max)


def group_median(x, group):
    return _group(x, group, np.median)


def group_min(x, group):
    return _group(x, group, np.min)


# def _group(x, group, agg_func):
#     if np.sum(np.isnan(x)) + np.sum(np.isnan(group)) > 0:
#         nan_mask = np.isnan(x) | np.isnan(group)
#
#         x_copy = np.where(nan_mask, 0, x)
#         group_copy = np.where(nan_mask, '0', group)
#
#         labels, indices = np.unique(group_copy, return_inverse=True)
#         grouped_val = np.array([agg_func(x_copy[group_copy == label]) for label in labels])
#         rlt = grouped_val[indices]
#         rlt = np.where(nan_mask, np.nan, rlt)
#     else:
#         labels, indices = np.unique(group, return_inverse=True)
#         grouped_val = np.array([agg_func(x[group == label]) for label in labels])
#         rlt = grouped_val[indices]
#     return rlt


def _group_neutralize_single(x, group):
    # Ensure x and group have same length
    assert len(x) == len(group), "Series x and group must have same length."

    # Calculate the mean of each group
    group_means = x.groupby(group).transform('mean')
    # Subtract the mean of each group from the corresponding elements in x
    neutralized_values = x - group_means

    return neutralized_values


def group_neutralize(x, group):
    return sub(x, group_mean(x, group))


def group_rank(x, group):
    pass


def group_sum(x, group):
    return _group(x, group, np.sum)


def group_zscore(x, group):
    pass


# Transformational Operators

def arc_cos(x):
    return np.arccos(x)


def arc_sin(x):
    return np.arc_sin(x)


def arc_tan(x):
    return np.arctan(x)


# def bucket(rank(x), range="0, 1, 0.1" or buckets = "2,5,6,7,10")

# def clamp(x, lower = 0, upper = 0, inverse = False, mask = ")

def filter(x, h = "1, 2, 3, 4", t="0.5"):
    pass


def keep(x, f, period=5):
    pass


def left_tail(x, maximum=0.02):
    """
    left_tail(rank(close), maximum = 0.02)
    """
    return if_else(more(x, maximum), np.nan, x)


def right_tail(x, minimum=0.5):
    """
    right_tail(rank(volume), minimum = 0.5)
    """
    return if_else(less(x, minimum), np.nan, x)


def sigmoid(x):
    return divide(1, add(1, np.exp(-x)))


def tanh(x):
    return np.tanh(x)


def tail(x, lower=0, upper=0.5, newval=np.nan):
    cond = logical_or(less(x, lower), more(x, upper))
    return if_else(cond, x, np.nan)


# TODO: 保持变量类型封闭, 目前有一堆BUG
# BUG:
def trade_when(trigger: np.ndarray[bool],
               alpha: np.ndarray[np.float64],
               exit_cond: np.ndarray[bool]):
    rlt = if_else(exit_cond, np.nan, alpha)
    rlt = if_else(trigger, rlt, ts_delay(rlt, 1))
    if isinstance(alpha, np.ndarray):
        return rlt
    elif isinstance(alpha, pd.Series):
        return pd.Series(rlt, index=alpha.index, name=alpha.name)
    elif isinstance(alpha, pd.DataFrame):
        return pd.DataFrame(rlt, index=alpha.index, columns=alpha.columns)



import pandas as pd
import numpy as np
import typing

def get_clean_factor_and_forward_returns(
    factor,
    prices,
    groupby=None,
    binning_by_group=False,
    quantiles=5,
    bins=None,
    periods=(1, 5, 20),
    periods_by_factor = False,
    filter_zscore=20,
    groupby_labels=None,
    zero_aware=False,
    cumulative_returns=True
):
    if periods_by_factor:
        periods = None
    forward_returns = compute_forward_returns(
        factor,
        prices,
        periods,
        periods_by_factor,
        filter_zscore,
        cumulative_returns,
    )

    factor_data = get_clean_factor(
        factor, 
        forward_returns, 
        groupby=groupby,
        groupby_labels=groupby_labels,
        quantiles=quantiles, 
        bins=bins,
        binning_by_group=binning_by_group,
        zero_aware=zero_aware)
    factor_data["factor_quantile"] = factor_data["factor_quantile"].astype(int)
    return factor_data

def get_clean_factor_and_current_returns(
    factor: pd.DataFrame,
    prices: pd.DataFrame,
    groupby: typing.Optional[typing.Union[pd.Series, dict]]=None,
    binning_by_group: bool=False,
    quantiles: typing.Union[int, typing.Iterable[float]]=5,
    bins=None,
    filter_zscore=20,
    groupby_labels=None,
    zero_aware=False,
):

    forward_returns = compute_current_returns(
        factor,
        prices,
    )

    factor_data = get_clean_factor(
        factor, 
        forward_returns, 
        groupby=groupby,
        groupby_labels=groupby_labels,
        quantiles=quantiles, 
        bins=bins,
        binning_by_group=binning_by_group,
        zero_aware=zero_aware)
    return factor_data


def compute_forward_returns_custom(
    factor: pd.DataFrame,
    prices,
    filter_zscore = None,
    cumlative_returns = False,
):
    factor_dateindex = factor.index.levels[0]
    factor_dateindex = factor_dateindex.intersection(prices.index)

    if len(factor_dateindex) == 0:
        raise ValueError("Factor and prices indices don't match")
    raise NotImplementedError("this function has not been implemented yet")


def compute_forward_returns(
    factor: pd.DataFrame,
    prices: pd.DataFrame,
    periods: tuple = (1, ),
    periods_by_factor: bool = False,
    filter_zscore: float = None,
    cumulative_returns: bool = False
) -> pd.DataFrame:
    """create the N period forward returns based on the trading prices

    Parameters
    ----------
    factor : pd.DataFrame
        DataFrame with multiIndex
    prices : pd.DataFrame

    periods : tuple, optional
        _description_, by default (1, )
    filter_zscore : float, optional
        _description_, by default None
    cumulative_returns : bool, optional
        _description_, by default False

    Returns
    -------
    pd.DataFrame
    """
    factor_dateindex = factor.index.get_level_values(0).unique()
    factor_dateindex = factor_dateindex.intersection(prices.index)

    if len(factor_dateindex) == 0:
        raise ValueError("Factor and prices indices don't match: make sure "
                         "they have the same convention in terms of datetimes "
                         "and symbol-names")

    raw_values_dict = {}
    column_list = []

    if not periods_by_factor:
        for period in sorted(periods):
            if cumulative_returns:
                returns = prices.pct_change(period)
            else:
                returns = prices.pct_change()

            forward_returns = returns.shift(-period).reindex(factor_dateindex)

            if filter_zscore is not None:
                mask = abs(forward_returns - forward_returns.mean()) > (filter_zscore * forward_returns.std())
                forward_returns[mask] = np.nan

            label = f"{period}D"

            column_list.append(label)

            raw_values_dict[label] = np.concatenate(forward_returns.values)
    else:
        returns = prices.loc[factor_dateindex].pct_change().fillna(0)
        forward_returns = returns.shift(-1).reindex(factor_dateindex)

        label = "1T"
        column_list.append(label)

        raw_values_dict[label] = np.concatenate(forward_returns.values)

    df = pd.DataFrame.from_dict(raw_values_dict)
    df.set_index(
        pd.MultiIndex.from_product(
            [factor_dateindex, prices.columns],
            names=['date', 'asset']
        ),
        inplace=True
    )
    df = df.reindex(factor.index)

    df = df[column_list]

    df.index.set_names(['date', 'asset'], inplace=True)
    return df

def compute_forward_returns_custom(
    factor: pd.DataFrame,
    prices,
    filter_zscore = None,
    cumlative_returns = False,
):
    factor_dateindex = factor.index.levels[0]
    factor_dateindex = factor_dateindex.intersection(prices.index)

    if len(factor_dateindex) == 0:
        raise ValueError("Factor and prices indices don't match")
    raise NotImplementedError("this function has not been implemented yet")


def compute_current_returns(
    factor: pd.DataFrame,
    prices: pd.DataFrame,
) -> pd.DataFrame:

    factor_dateindex = factor.index.get_level_values(0).unique()
    factor_dateindex = factor_dateindex.intersection(prices.index)

    if len(factor_dateindex) == 0:
        raise ValueError("Factor and prices indices don't match: make sure "
                         "they have the same convention in terms of datetimes "
                         "and symbol-names")

    raw_values_dict = {}
    column_list = []


    returns = prices.pct_change().reindex(factor_dateindex)

    label = "current_returns"

    column_list.append(label)

    raw_values_dict[label] = np.concatenate(returns.values)

    df = pd.DataFrame.from_dict(raw_values_dict)
    df.set_index(
        pd.MultiIndex.from_product(
            [factor_dateindex, prices.columns],
            names=['date', 'asset']
        ),
        inplace=True
    )
    df = df.reindex(factor.index)
    df = df[column_list]
    df.index.set_names(['date', 'asset'], inplace=True)
    return df




def quantize_factor(factor_data,
                    quantiles=5,
                    bins=None,
                    by_group=False,
                    no_raise=False,
                    zero_aware=False):
    """
    Computes period wise factor quantiles.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.

        - See full explanation in utils.get_clean_factor_and_forward_returns

    quantiles : int or sequence[float]
        Number of equal-sized quantile buckets to use in factor bucketing.
        Alternately sequence of quantiles, allowing non-equal-sized buckets
        e.g. [0, .10, .5, .90, 1.] or [.05, .5, .95]
        Only one of 'quantiles' or 'bins' can be not-None
    bins : int or sequence[float]
        Number of equal-width (valuewise) bins to use in factor bucketing.
        Alternately sequence of bin edges allowing for non-uniform bin width
        e.g. [-4, -2, -0.5, 0, 10]
        Only one of 'quantiles' or 'bins' can be not-None
    by_group : bool, optional
        If True, compute quantile buckets separately for each group.
    no_raise: bool, optional
        If True, no exceptions are thrown and the values for which the
        exception would have been thrown are set to np.NaN
    zero_aware : bool, optional
        If True, compute quantile buckets separately for positive and negative
        signal values. This is useful if your signal is centered and zero is
        the separation between long and short signals, respectively.

    Returns
    -------
    factor_quantile : pd.Series
        Factor quantiles indexed by date and asset.
    """
    if not ((quantiles is not None and bins is None) or
            (quantiles is None and bins is not None)):
        raise ValueError('Either quantiles or bins should be provided')

    if zero_aware and not (isinstance(quantiles, int)
                           or isinstance(bins, int)):
        msg = ("zero_aware should only be True when quantiles or bins is an"
               " integer")
        raise ValueError(msg)

    def quantile_calc(x, _quantiles, _bins, _zero_aware, _no_raise):
        try:
            if _quantiles is not None and _bins is None and not _zero_aware:
                return pd.qcut(x, _quantiles, labels=False) + 1
            elif _quantiles is not None and _bins is None and _zero_aware:
                pos_quantiles = pd.qcut(x[x >= 0], _quantiles // 2,
                                        labels=False) + _quantiles // 2 + 1
                neg_quantiles = pd.qcut(x[x < 0], _quantiles // 2,
                                        labels=False) + 1
                return pd.concat([pos_quantiles, neg_quantiles]).sort_index()
            elif _bins is not None and _quantiles is None and not _zero_aware:
                return pd.cut(x, _bins, labels=False) + 1
            elif _bins is not None and _quantiles is None and _zero_aware:
                pos_bins = pd.cut(x[x >= 0], _bins // 2,
                                  labels=False) + _bins // 2 + 1
                neg_bins = pd.cut(x[x < 0], _bins // 2,
                                  labels=False) + 1
                return pd.concat([pos_bins, neg_bins]).sort_index()
        except Exception as e:
            if _no_raise:
                return pd.Series(np.nan, index=x.index)
            raise e

    grouper = [factor_data.index.get_level_values('date')]
    if by_group:
        grouper.append('group')

    factor_quantile = factor_data.groupby(grouper, group_keys=False)['factor'].apply(quantile_calc, quantiles, bins, zero_aware, no_raise)
    factor_quantile.name = 'factor_quantile'

    return factor_quantile.dropna()


def get_clean_factor(factor,
                     forward_returns,
                     groupby=None,
                     binning_by_group=False,
                     quantiles=5,
                     bins=None,
                     groupby_labels=None,
                     zero_aware=False):
    """
    Formats the factor data, forward return data, and group mappings into a
    DataFrame that contains aligned MultiIndex indices of timestamp and asset.
    The returned data will be formatted to be suitable for Alphalens functions.

    It is safe to skip a call to this function and still make use of Alphalens
    functionalities as long as the factor data conforms to the format returned
    from get_clean_factor_and_forward_returns and documented here

    Parameters
    ----------
    factor : pd.Series - MultiIndex
        A MultiIndex Series indexed by timestamp (level 0) and asset
        (level 1), containing the values for a single alpha factor.
        ::
            -----------------------------------
                date    |    asset   |
            -----------------------------------
                        |   AAPL     |   0.5
                        -----------------------
                        |   BA       |  -1.1
                        -----------------------
            2014-01-01  |   CMG      |   1.7
                        -----------------------
                        |   DAL      |  -0.1
                        -----------------------
                        |   LULU     |   2.7
                        -----------------------

    forward_returns : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by timestamp (level 0) and asset
        (level 1), containing the forward returns for assets.
        Forward returns column names must follow the format accepted by
        pd.Timedelta (e.g. '1D', '30m', '3h15m', '1D1h', etc).
        'date' index freq property must be set to a trading calendar
        (pandas DateOffset), see infer_trading_calendar for more details.
        This information is currently used only in cumulative returns
        computation
        ::
            ---------------------------------------
                       |       | 1D  | 5D  | 10D
            ---------------------------------------
                date   | asset |     |     |
            ---------------------------------------
                       | AAPL  | 0.09|-0.01|-0.079
                       ----------------------------
                       | BA    | 0.02| 0.06| 0.020
                       ----------------------------
            2014-01-01 | CMG   | 0.03| 0.09| 0.036
                       ----------------------------
                       | DAL   |-0.02|-0.06|-0.029
                       ----------------------------
                       | LULU  |-0.03| 0.05|-0.009
                       ----------------------------

    groupby : pd.Series - MultiIndex or dict
        Either A MultiIndex Series indexed by date and asset,
        containing the period wise group codes for each asset, or
        a dict of asset to group mappings. If a dict is passed,
        it is assumed that group mappings are unchanged for the
        entire time period of the passed factor data.
    binning_by_group : bool
        If True, compute quantile buckets separately for each group.
        This is useful when the factor values range vary considerably
        across gorups so that it is wise to make the binning group relative.
        You should probably enable this if the factor is intended
        to be analyzed for a group neutral portfolio
    quantiles : int or sequence[float]
        Number of equal-sized quantile buckets to use in factor bucketing.
        Alternately sequence of quantiles, allowing non-equal-sized buckets
        e.g. [0, .10, .5, .90, 1.] or [.05, .5, .95]
        Only one of 'quantiles' or 'bins' can be not-None
    bins : int or sequence[float]
        Number of equal-width (valuewise) bins to use in factor bucketing.
        Alternately sequence of bin edges allowing for non-uniform bin width
        e.g. [-4, -2, -0.5, 0, 10]
        Chooses the buckets to be evenly spaced according to the values
        themselves. Useful when the factor contains discrete values.
        Only one of 'quantiles' or 'bins' can be not-None
    groupby_labels : dict
        A dictionary keyed by group code with values corresponding
        to the display name for each group.
    max_loss : float, optional
        Maximum percentage (0.00 to 1.00) of factor data dropping allowed,
        computed comparing the number of items in the input factor index and
        the number of items in the output DataFrame index.
        Factor data can be partially dropped due to being flawed itself
        (e.g. NaNs), not having provided enough price data to compute
        forward returns for all factor values, or because it is not possible
        to perform binning.
        Set max_loss=0 to avoid Exceptions suppression.
    zero_aware : bool, optional
        If True, compute quantile buckets separately for positive and negative
        signal values. This is useful if your signal is centered and zero is
        the separation between long and short signals, respectively.
        'quantiles' is None.

    Returns
    -------
    merged_data : pd.DataFrame - MultiIndex
        A MultiIndex Series indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.

        - forward returns column names follow the format accepted by
          pd.Timedelta (e.g. '1D', '30m', '3h15m', '1D1h', etc)

        - 'date' index freq property (merged_data.index.levels[0].freq) is the
          same as that of the input forward returns data. This is currently
          used only in cumulative returns computation
        ::
           -------------------------------------------------------------------
                      |       | 1D  | 5D  | 10D  |factor|group|factor_quantile
           -------------------------------------------------------------------
               date   | asset |     |     |      |      |     |
           -------------------------------------------------------------------
                      | AAPL  | 0.09|-0.01|-0.079|  0.5 |  G1 |      3
                      --------------------------------------------------------
                      | BA    | 0.02| 0.06| 0.020| -1.1 |  G2 |      5
                      --------------------------------------------------------
           2014-01-01 | CMG   | 0.03| 0.09| 0.036|  1.7 |  G2 |      1
                      --------------------------------------------------------
                      | DAL   |-0.02|-0.06|-0.029| -0.1 |  G3 |      5
                      --------------------------------------------------------
                      | LULU  |-0.03| 0.05|-0.009|  2.7 |  G1 |      2
                      --------------------------------------------------------
    """

    # initial_amount = float(len(factor.index))

    factor_copy = factor.copy()
    factor_copy.index = factor_copy.index.rename(['date', 'asset'])
    factor_copy = factor_copy[np.isfinite(factor_copy)]

    merged_data = forward_returns.copy()
    merged_data['factor'] = factor_copy

    if groupby is not None:
        if isinstance(groupby, dict):
            diff = set(factor_copy.index.get_level_values(
                'asset')) - set(groupby.keys())
            if len(diff) > 0:
                raise KeyError(
                    "Assets {} not in group mapping".format(
                        list(diff)))

            ss = pd.Series(groupby)
            groupby = pd.Series(index=factor_copy.index,
                                data=ss[factor_copy.index.get_level_values(
                                    'asset')].values)

        if groupby_labels is not None:
            diff = set(groupby.values) - set(groupby_labels.keys())
            if len(diff) > 0:
                raise KeyError(
                    "groups {} not in passed group names".format(
                        list(diff)))

            sn = pd.Series(groupby_labels)
            groupby = pd.Series(index=groupby.index,
                                data=sn[groupby.values].values)

        merged_data['group'] = groupby.astype('category')

    merged_data = merged_data.dropna()

    # fwdret_amount = float(len(merged_data.index))

    # no_raise = False if max_loss == 0 else True

    quantile_data = quantize_factor(
        merged_data,
        quantiles,
        bins,
        binning_by_group,
        True,
        zero_aware
    )

    merged_data['factor_quantile'] = quantile_data

    merged_data = merged_data.dropna()

    # binning_amount = float(len(merged_data.index))

    # tot_loss = (initial_amount - binning_amount) / initial_amount
    # fwdret_loss = (initial_amount - fwdret_amount) / initial_amount
    # bin_loss = tot_loss - fwdret_loss

    # print("Dropped %.1f%% entries from factor data: %.1f%% in forward "
    #       "returns computation and %.1f%% in binning phase "
    #       "(set max_loss=0 to see potentially suppressed Exceptions)." %
    #       (tot_loss * 100, fwdret_loss * 100, bin_loss * 100))

    # if tot_loss > max_loss:
    #     message = ("max_loss (%.1f%%) exceeded %.1f%%, consider increasing it."
    #                % (max_loss * 100, tot_loss * 100))
    #     raise MaxLossExceededError(message)
    # else:
    #     print("max_loss is %.1f%%, not exceeded: OK!" % (max_loss * 100))

    return merged_data


import pandas as pd
import sys
import io
import pyperclip

def copy_to_clipboard(func):
    def wrapper(*args, **kwargs):
        # 使用StringIO对象来捕获print输出
        output = func(*args, **kwargs)
        
        # 将捕获的输出复制到剪切板
        pyperclip.copy(output)

    return wrapper

def df_to_latex(df: pd.DataFrame, file, caption=None, label=None):
    result = df.style.to_latex(
        hrules=True,
        position="htbp",
        position_float="centering",
        caption=caption, 
        label=label
    )
    with open(file, "w+") as f:
        f.write(result)

def df_to_html(df, caption=None):
    n_cols = len(df.columns) + 1
    html = df.to_html()
    # 定义前缀和后缀
    prefix = f"""\
<table style="border-collapse: collapse; width: 100%;">
  <caption style="padding: 10px; font-weight: bold;">{caption}</caption>"""

    suffix = """\
  <tfoot>
      <tr style="text-align: right;">
          <td colspan="6" style="border-top: 2px solid black;"></td>
      </tr>
  </tfoot>
</table>
"""

    # 替换原始表格的前缀和后缀
    html = html.replace('<table border="1" class="dataframe">', prefix).replace('</table>', suffix)

    # 为标题行的<th>标签添加上下边框
    html = html.replace('<th>', '<th style="border-top: 2px solid black; border-bottom: 2px solid black; padding: 5px 10px;">', n_cols)  # 只替换前6次，对应标题行的6列

    # 为表体的<tr>标签添加右对齐属性
    html = html.replace('<tr>', '<tr style="text-align: right;">')
    html = html.replace('<th>', '<th style="padding: 5px 10px;">')  # 替换其余的<th>标签，也即是表体中的索引列
    html = html.replace('<td>', '<td style="padding: 5px 10px;">')  # 给<td>标签添加样式

    # 返回转换后的HTML
    return html

@copy_to_clipboard
def df_to_html_to_clipboard(df, caption=None):
    return df_to_html(df, caption)





raise NotImplementedError

class Pipeline:
    def __init__(self, context):
        self.context = context
        self.trade_dates = TRADE_DATES
        self._date_iter = iter(TRADE_DATES)
        self.current_date = None
        self._results = []

    def next(self):
        self.current_date = next(self._date_iter)

    def order_pct_to(self, target_percent):
        #TODO: add type checking
        if target_percent is not None:
            self._results.append(pd.Series(target_percent, index=self.universe, name=self.current_date))

    def initialize(self):
        self.universe = SYMBOLS

    def on_session_end(self):
        raise NotImplementedError("on_session_end should be implemented in subclasses")
        if self.current_date in REBALANCE_DATES:
            _mu, _sigma, _ = self.context.query_covariance(self.current_date, 50)
            if _mu is not None:
                _target_positions = self.strategy(5, 10)(_mu, _sigma)
            else:
                _target_positions = np.repeat(1/5, 5) 
            self.order_pct_to(_target_positions)
    
    def _on_session_end(self):
        self.on_session_end()
    
    def run_pipeline(self):
        self.initialize()
        while True:
            try:
                self.next()
                self._on_session_end()
            except StopIteration:
                break
        return pd.concat(self._results, axis=1).T

class MinVarPipe(Pipeline):
    def initialize(self):
        self.universe = SYMBOLS

    def on_session_end(self):
        if self.current_date in REBALANCE_DATES:
            _mu, _sigma, _ = self.context.query_covariance(self.current_date, 50)
            if _mu is not None:
                _target_positions = MinVarPortfolio(5)(_mu, _sigma)
            else:
                _target_positions = np.repeat(1/5, 5)
            self.order_pct_to(_target_positions)

class MeanVarPipe(Pipeline):
    def initialize(self):
        self.universe = SYMBOLS

    def on_session_end(self):
        if self.current_date in REBALANCE_DATES:
            _mu, _sigma, _ = self.context.query_covariance(self.current_date, 50)
            if _mu is not None:
                _target_positions = MeanVarPortfolio(5, 0.1, _mu)(_mu, _sigma)
            else:
                _target_positions = np.repeat(1/5, 5)

            self.order_pct_to(_target_positions)

class QUPipe(Pipeline):
    def initialize(self):
        self.universe = SYMBOLS

    def on_session_end(self):
        if self.current_date in REBALANCE_DATES:
            _mu, _sigma, _ = self.context.query_covariance(self.current_date, 50)
            if _mu is not None:
                _target_positions = QuadUtilityPortfolio(5, 20)(_mu, _sigma)
            else:
                _target_positions = np.repeat(1/5, 5)

            self.order_pct_to(_target_positions)


import pandas as pd
import numpy as np
import math
from .calendars import DEFAULT_CALENDAR
from typing import Iterable, Literal


class QuickFactorTestor:
    def __init__(self, start_date, end_date, prices):
        self.universe = list(prices.columns)
        self.n_assets = len(prices.columns)
        self._model = QuickBackTestor(start_date, end_date, prices) 

    def evaluate(self, 
        cleaned_factor: pd.DataFrame, 
        type: Literal["long_and_short", "group"]="long_and_short", 
    ):
        """_summary_

        Parameters
        ----------
        """
        self._model.reset()

        if type == "long_and_short":
            target_positions = self.get_long_short_positions(cleaned_factor)
            return self._model.run_backtest(target_positions)
        elif type == "group":
            n_groups = cleaned_factor["factor_quantile"].max()
            target_positions_list = [self.get_group_positions(cleaned_factor, i+1) for i in range(n_groups)]
            return self._model.run_backtests(target_positions_list)
    
    def get_group_positions(self, cleaned_factor: pd.DataFrame, factor_quantile: int):
        tmp = cleaned_factor.query(f"factor_quantile=={factor_quantile}")["factor_quantile"].unstack().reindex(self.universe, axis=1)
        # tmp = cleaned_factor[cleaned_factor['factor_quantile'] == factor_quantile]["factor_quantile"].unstack().reindex(self.universe, axis=1)
        weights = (tmp > 0) / np.repeat((tmp > 0).sum(axis=1).values.reshape(-1, 1), self.n_assets, axis=1)
        return weights

    def get_long_short_positions(self, cleaned_factor: pd.DataFrame):
        max_quantile = cleaned_factor["factor_quantile"].max()
        min_quantile = cleaned_factor["factor_quantile"].min()
        long_positions = self.get_group_positions(cleaned_factor, max_quantile)
        short_positions = self.get_group_positions(cleaned_factor, min_quantile)
        return long_positions - short_positions

    def summary(self):
        pass


class QuickBackTestor:
    def __init__(self, start_date, end_date, prices):
        self.start_date = start_date
        self.end_date = end_date
        self._prices = prices
        self._returns = self._prices.pct_change().fillna(0)
        self._commission = None
        self._trade_dates = DEFAULT_CALENDAR.sessions_in_range(self.start_date, self.end_date)
        self._results = {"strategy_rewards": None}
 
    def run_backtest(self, strategy: pd.DataFrame):
        if not set(strategy.loc[self.start_date:self.end_date].index).issubset(set(self._trade_dates)):
            raise ValueError("the index of strategy must be a trade date")
        
        if strategy.isna().sum().sum() > 0:
            raise ValueError("the elements of startegy should not be nan")
        
        strategy = strategy.loc[self.start_date:self.end_date]
        rebalance_dates = strategy.index.to_list()
        dates = list(zip(rebalance_dates[:-1], rebalance_dates[1:]))
        results = []
        for start_date, end_date in dates:
            result = self._rewards_in_range(strategy.loc[start_date], start_date, end_date)
            results.append(result)
        self._results["strategy_rewards"] = pd.concat(results).reindex(self._trade_dates).fillna(0)
        return self._results["strategy_rewards"]
    
    def run_backtests(self, strategies: Iterable[pd.DataFrame]):
        """run multiple backtests
        The strategies should be aligned in the same format

        Parameters
        ----------
        strategies : Iterable[pd.DataFrame]
            _description_
        """
        sample_strategy = strategies[0]
        if not set(sample_strategy.index).issubset(set(self._trade_dates)):
            raise ValueError("the index of strategy must be a trade date")
        
        if sample_strategy.isna().sum().sum() > 0:
            raise ValueError("the elements of startegy should not be nan")
        
        rebalance_dates = sample_strategy.index.to_list()
        dates = list(zip(rebalance_dates[:-1], rebalance_dates[1:]))
        results = [] 
        for start_date, end_date in dates:
            current_strategy = pd.concat([strategy.loc[start_date] for strategy in strategies], axis=1, ignore_index=True)
            result = self._rewards_in_range(current_strategy, start_date, end_date)
            results.append(result)
        self._results["strategy_rewards"] = pd.concat(results).reindex(self._trade_dates).fillna(0)  
        return self._results["strategy_rewards"]

    def reset(self):
        self._results = {"strategy_rewards": None}

    def summary(self):
        pass

    def _rewards_in_range(self, start_position, start_date, end_date):
        returns_slice = self._returns.loc[start_date:end_date].copy(deep=True)
        # construct the imaginary returns:
        # short a stock is equal to long the stock that is purely negatively correlated with the prime stock
        returns_slice.iloc[:, np.where(start_position < 0)[0]] *= -1 
        start_position = np.abs(start_position)
        returns_slice.iloc[0] = 0
        pnls = (returns_slice + 1).cumprod()
        pnl = pnls.dot(start_position)
        rewards = pnl.pct_change()
        return rewards.iloc[1:]
    

class SlowBackTestor:
    def __init__(self, start_date, end_date, prices):
        self.start_date = start_date
        self.end_date = end_date
        self._prices = prices
        self._returns = self._prices.pct_change().fillna(0)
        self._commission = None
        self._trade_dates = DEFAULT_CALENDAR.sessions_in_range(self.start_date, self.end_date)
        self._results = {"strategy_rewards": None}
 
    def run_backtest(self, strategy: pd.DataFrame):
        if not set(strategy.loc[self.start_date:self.end_date].index).issubset(set(self._trade_dates)):
            raise ValueError("the index of strategy must be a trade date")
        
        if strategy.isna().sum().sum() > 0:
            raise ValueError("the elements of startegy should not be nan")
        
        strategy = strategy.loc[self.start_date:self.end_date]
        rebalance_dates = strategy.index.to_list()
        dates = list(zip(rebalance_dates[:-1], rebalance_dates[1:]))
        results = []
        for start_date, end_date in dates:
            target_position = strategy.loc[start_date]
            # masking stocks that can not be traded
            result = self._rewards_in_range(target_position, start_date, end_date)
            results.append(result)
        self._results["strategy_rewards"] = pd.concat(results).reindex(self._trade_dates).fillna(0)
        return self._results["strategy_rewards"]
    
    def run_backtests(self, strategies: Iterable[pd.DataFrame]):
        """run multiple backtests
        The strategies should be aligned in the same format

        Parameters
        ----------
        strategies : Iterable[pd.DataFrame]
            _description_
        """
        sample_strategy = strategies[0]
        if not set(sample_strategy.index).issubset(set(self._trade_dates)):
            raise ValueError("the index of strategy must be a trade date")
        
        if sample_strategy.isna().sum().sum() > 0:
            raise ValueError("the elements of startegy should not be nan")
        
        rebalance_dates = sample_strategy.index.to_list()
        dates = list(zip(rebalance_dates[:-1], rebalance_dates[1:]))
        results = [] 
        for start_date, end_date in dates:
            current_strategy = pd.concat([strategy.loc[start_date] for strategy in strategies], axis=1, ignore_index=True)
            result = self._rewards_in_range(current_strategy, start_date, end_date)
            results.append(result)
        self._results["strategy_rewards"] = pd.concat(results).reindex(self._trade_dates).fillna(0)  
        return self._results["strategy_rewards"]

    def reset(self):
        self._results = {"strategy_rewards": None}

    def summary(self):
        pass

    def _rewards_in_range(self, start_position, start_date, end_date):
        returns_slice = self._returns.loc[start_date:end_date].copy(deep=True)
        # construct the imaginary returns:
        # short a stock is equal to long the stock that is purely negatively correlated with the prime stock
        returns_slice.iloc[:, np.where(start_position < 0)[0]] *= -1 
        start_position = np.abs(start_position)
        returns_slice.iloc[0] = 0
        pnls = (returns_slice + 1).cumprod()
        pnl = pnls.dot(start_position)
        rewards = pnl.pct_change()
        return rewards.iloc[1:]


raise DeprecationWarning("the frame class is deprecated and will be removed")
import pickle
from typing import Any
import numpy as np
import pandas as pd
import copy
from .calendars import XSHGExchangeCalendar
import os
import itertools
import pickle

# os.path.join(os.path.dirname(__file__))

# with open(f"{os.path.dirname(__file__)}/data/data.pkl", "rb") as f:
#     trade_dates, all_symbol, all_datas = pickle.load(f)

def backward_match(data, match):
    return np.searchsorted(data, match, "left")

def forward_match(data, match):
    return np.searchsorted(data, match, "right") - 1

NO_ACTION = frozenset("NO_ACTION")


# class DataPortal:
#     def __init__(self, trade_dates=None, symbols=None, fields=None, data=None):
#         self._times = trade_dates
#         self._symbols = np.array(all_symbol)
#         self._fields = ["open", "close"]

#         self._hash_times = pd.Series(range(len(self._times)), index=self._times)
#         self._hash_symbols = pd.Series(range(len(self._symbols)), index=self._symbols)
#         self._hash_fields = pd.Series(range(len(self._fields)), index=self._fields)

#         self._values = all_datas
#         self._values.flags.writeable = False  # read only values

#     def get_data_at_date(self, date, deep_copy=False):
#         idx = self._times.searchsorted(date, side='right') - 1
#         if not deep_copy:
#             return self._values[idx], self._times[idx]
#         else:
#             return copy.deepcopy(self._values[idx]), self._times[idx]
         
#     def get_data_at_symbol(self, symbol, deep_copy=False):
#         idx = self._symbols.searchsorted(symbol)
#         return self._values[:, idx, :]
    
#     def get_data_at_field(self, field, deep_copy=False):
#         idx = self._fields.index(field)
#         return self._values[:, :, idx]
    
#     def history(self, date, bar_count, symbols):
#         idx = self._times.searchsorted(date, side='right') - 1
#         return self._values[idx-bar_count+1: idx+1][:, self._hash_symbols[symbols], :]
    
#     def hash_history(self, date, bar_count=1):
#         return self._values[self._hash_times[date]]

#     def create_subenv(self, trade_dates: list, symbols: list, fileds: list):
#         ret_times = self._hash_times[trade_dates]
#         ret_symbols = self._hash_symbols[symbols]
#         ret_fields = self._hash_fields[fileds]
        
#         return self._values[ret_times.values, :, :][:, ret_symbols.values, :][:,:, ret_fields], ret_times.index, ret_symbols.index, ret_fields.index


class BaseFrame:
    
    def history(self, data, bar_count):
        pass

class NDFrame:
    def __init__(self, data, times, symbols, fields):
        self._values = data
        self._times = list(times)
        self._symbols = symbols
        self._fields = fields

        self._hash_times = pd.Series(range(len(self._times)), index=self._times)
        self._hash_symbols = pd.Series(range(len(self._symbols)), index=self._symbols)
        self._hash_fields = pd.Series(range(len(self._fields)), index=self._fields)
        self._values.flags.writeable = False  # read only values

    def history(self, date, bar_acount=1):
        pass

    def __len__(self):
        return len(self._values)

    def fillna(self):
        pass
    
    def save(self, file):
        with open(file, "wb") as f:
            pickle.dump(self, f)

    @property
    def symbols(self):
        return self._symbols
    
    @property
    def trade_dates(self):
        return self._times

    @property
    def fields(self):
        return self._fields

    @classmethod
    def load(cls, file) -> "NDFrame":
        with open(file, "rb") as f:
            ndf = pickle.load(f)
        return ndf
    
    def drop_to_DataFrame(self, field):
        data = self._values[:,:,self._hash_fields[field]]
        return pd.DataFrame(data=data, index=self._times, columns=self._symbols)
    

class DataIter:
    def __init__(self, data: NDFrame):
        self.data = data
        self.current_idx = 0
    
    def reset(self):
        self.current_idx = 0

    def next(self):
        self.current_idx += 1
        if self.current_idx > len(self.data) - 2:
            raise StopIteration
        
    def spot_price(self, symbol=None):
        return self.data._values[self.current_idx, :, self.data._hash_fields["close"]] 

    @property
    def current_data(self):
        return self.data._values[self.current_idx]

    @property
    def current_date(self):
        return self.data._times[self.current_idx]

# class DataIter:
#     def __init__(self, start_date, end_date, data_portal: DataPortal):
#         self.start_date = start_date
#         self.end_date = end_date
#         # self.trade_dates = trade_dates
#         self.true_index = 0
#         self.current_index = 0
#         self.end_index = forward_match(self.trade_dates, end_date)
#         # self.dirty = True
#         self.data_portal = data_portal
#         self.reset(start_date)
    
#     def reset(self, date):
#         self.date = date
#         self.current_index = 0
#         self.true_index = backward_match(self.trade_dates, date)

#     def iter(self):
#         return self

#     def next(self):
#         self.true_index += 1
#         self.current_index += 1
#         if self.true_index > self.end_index:
#             self.true_index -= 1
#             self.current_index -= 1
#             raise StopIteration
    
#     def spot(self, symbol=None, field=None):
#         return self.data_portal._values[self.true_index]
    
#     def spot_price(self, symbol=None):
#         return self.data_portal._values[self.true_index, :, self.data_portal._hash_fields["close"]] 

#     @property
#     def current_data(self):
#         return self.data_portal._values[self.true_index]

#     @property
#     def current_date(self):
#         return self.trade_dates[self.true_index]


import pandas as pd
import numpy as np
import statsmodels.api as sm
from statsmodels.regression.rolling import RollingOLS


# def rolling_regression(endog, exog, window: int):
#     rols = RollingOLS(endog, exog, window=window)
#     rres = rols.fit()
#     aa = rres.params.copy()
#     aa = aa.reset_index()
#     aa["asset"] = symbol
#     aa = aa.rename({"index": "date"}, axis=1)
#     aa.set_index(["date", "asset"], inplace=True)
#     return aa


# -*- coding: utf-8 -*-
from alphagens.data_source._dataapi_linux36 import Client
import pandas as pd
from io import StringIO

TOKEN = "8d08961ec3c2ff4f22b627f976ed1583d792247526e6cde96e98d8cdd645a259"

def split_date_range(start_date, end_date, n_splits):
    start_date = pd.to_datetime(start_date)
    end_date = pd.to_datetime(end_date)
    total_days = (end_date - start_date).days
    days_per_split = total_days // n_splits
    
    date_ranges = []
    current_start = start_date
    for _ in range(n_splits - 1):
        current_end = current_start + timedelta(days=days_per_split - 1)
        date_ranges.append((current_start.strftime('%Y%m%d'), current_end.strftime('%Y%m%d')))
        current_start = current_end + timedelta(days=1)
    date_ranges.append((current_start.strftime('%Y%m%d'), end_date.strftime('%Y%m%d')))
    
    return date_ranges
			

def EquIndustryGet(secID="",ticker=["688005", "600001"], industryVersionCD="010303", into_date="", industry="", industryID="", industryID1=""):
	try:
		client = Client()
		client.init(TOKEN)
		url2=f'/api/equity/getEquIndustry.csv?\
field=&industryVersionCD={industryVersionCD}\
&industry=\
&secID=\
&ticker={",".join(ticker)}\
&intoDate={into_date}\
&industryID1=\
&industryID2=\
&industryID3=\
&equTypeID=\
&industryID='
		code, result = client.getData(url2)
		if(code==200):
			data = result.decode("gbk")
			df = pd.read_csv(StringIO(data))
			return df
		else:
			print (code)
			print (result)
	except Exception as e:
		raise e


def MktStockFactorsOneDayProGet(trade_date="20170612",secID="",ticker=["000001", "600000"], field=["STOM"],pandas="1"):
	try:
		client = Client()
		client.init(TOKEN)
		url2=f'/api/market/getMktStockFactorsOneDayPro.csv?\
field={",".join(field)}&\
secID=&\
ticker={",".join(ticker)}&\
tradeDate={trade_date}'
		code, result = client.getData(url2)
		if(code==200):
			data = result.decode("gbk")
			df = pd.read_csv(StringIO(data))
			return df
		else:
			print (code)
			print (result)
	except Exception as e:
		raise e


def MktStockFactorsOneDayGet(trade_date="20150227",secID="",ticker=["000001","600000"], field="",pandas="1"):
	try:
		client = Client()
		client.init(TOKEN)
		url2=f'/api/market/getStockFactorsOneDay.csv?\
field=&\
secID=&\
ticker={",".join(ticker)}&\
tradeDate={trade_date}'

		code, result = client.getData(url2)
		if(code==200):
			data = result.decode("gbk")
			df = pd.read_csv(StringIO(data))
			return df
		else:
			print (code)
			print (result)
	except Exception as e:
		raise e


def MktStockFactorsDateRangeGet(secID="", ticker=["000001", "000002"], start_date="20170612", end_date="20170616", field=[], pandas="1"):
	try:
		client = Client()
		client.init(TOKEN)
		url2=f'/api/market/getStockFactorsDateRange.csv?\
field=&\
secID=&\
ticker={",".join(ticker)}&\
beginDate={start_date}&\
endDate={end_date}'

		code, result = client.getData(url2)
		if(code==200):
			data = result.decode("gbk")
			df = pd.read_csv(StringIO(data))
			return df
		else:
			print (code)
			print (result)
	except Exception as e:
		raise e


# def P_MktStockFactorsDateRangeGet(secID="", ticker=["000001", "000002"], start_date="20170612", end_date="20170616", field=[], pandas="1", n_jobs=4, n_splits=8):
# 	inputs = None
# date_ranges = split_date_range(start_date, end_date, n_splits)	
#     results = Parallel(n_jobs=n_jobs)(
#         delayed(MktStockFactorsDateRangeGet)(secID="", ticker=ticker, start_date=s_date, end_date=e_date, field=field, pandas="1")
#         for s_date, e_date in date_ranges
#     )
# 	return results


if __name__ == '__main__':
	EquIndustryGet(ticker=["000001", "000002"])


from distutils.core import setup
from setuptools import find_packages
from Cython.Build import cythonize
import numpy
import glob
import platform

PLATFORM = platform.system()
from alphagens import __version__
# extensions = glob.glob("torchqtm/_C/*.pyx")
# extensions.extend(glob.glob("torchqtm/window/*.pyx"))
# extensions.extend(glob.glob("torchqtm/finance/*.pyx"))

install_requires = [
    'numpy',
    'pandas',
    'pandas_market_calendars',
]


def main():
    setup(
        # ext_modules=cythonize(extensions, annotate=True),
        include_dirs=[numpy.get_include()],
        name="alphagens",
        version=__version__,
        author="ny",
        author_email="nymath@163.com",
        install_requires=install_requires,
        description="An event-driven backtesting system",
        long_description=open('README.md', 'r').read(),
        long_description_content_type="text/markdown",
        download_url="https://github.com/nymath/alphagens/releases/tag/v1.0",
        packages=find_packages(),
        classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 3 - Alpha',
            'Programming Language :: Python :: 3',
            "License :: OSI Approved :: MIT License",
            "Operating System :: Unix",
        ],
        project_urls={
            "Homepage": "https://github.com/nymath/alphagens",
            'Documentation': 'https://github.com/nymath/alphagens',
            # 'Funding': 'https://donate.pypi.org',
            "Download": "https://github.com/nymath/alphagens/releases/tag/v1.0",
            'Source': 'https://github.com/nymath/alphagens',
            # 'Tracker': 'https://github.com/pypa/sampleproject/issues',
        },
        python_requires='>=3.9',
        entry_points={
        "console_scripts": [
            "texgen=alphagens.cli.texgen:main",
        ]
        }
    )


if __name__ == "__main__":
    main()

## setup cython
# python setup.py build_ext --inplace
## setup package
# pip install twine
# python setup.py sdist bdist_wheel
# twine upload dist/*
# twine upload --skip-existing dist/*

# ny.math
# xxx



from collections import OrderedDict
import pandas as pd
import datetime
from pandas_market_calendars import get_calendar as calendar_template
import typing
import numpy as np
from collections import defaultdict
DateLike = typing.Union[pd.Timestamp, str, int, float, datetime.datetime]

GLOBAL_DEFAULT_START = pd.Timestamp('2005-01-04')
GLOBAL_DEFAULT_END = pd.Timestamp.now().floor("D") + pd.DateOffset(years=1)

NANOSECONDS_PER_MINUTE = int(6e10)


def convert_date_to_int(dt):
    t = dt.year * 10000 + dt.month * 100 + dt.day
    t *= 1e6
    return t

def convert_dates_to_int(dates: pd.DatetimeIndex):
    intDates = dates.year * 10000 + dates.month * 100 + dates.day
    intDates *= 1e6
    return intDates.values.astype(np.uint64)


class TradingDatesMixin(object):
    def __init__(self, trading_calendars):
        pass


class XSHGExchangeCalendar(object):

    calendar = calendar_template('SSE')
    market_open_time = datetime.time(hour=9, minute=30)
    market_break_start_time = datetime.time(hour=11, minute=30)
    market_break_end_time = datetime.time(hour=13, minute=0)
    market_close_time = datetime.time(hour=15, minute=0)

    @classmethod
    def default_start(cls) -> pd.Timestamp:
        return GLOBAL_DEFAULT_START

    @classmethod
    def default_end(cls) -> pd.Timestamp:
        return GLOBAL_DEFAULT_END

    def __init__(self,
                 start_date: DateLike = None,
                 end_date: DateLike = None,
                 side: typing.Literal["left", "right", "both", "neither"] = "left"):
        self.side = side
        if start_date is None:
            start_date = self.default_start()

        if end_date is None:
            end_date = self.default_end()

        if start_date >= end_date:
            raise ValueError(
                "`start` must be earlier than `end` although `start` parsed as"
                f" '{start_date}' and `end` as '{end_date}'."
            )
        self.start_date = start_date
        self.end_date = end_date
        self.valid_days: pd.DatetimeIndex = pd.DatetimeIndex(self.calendar.valid_days(start_date, end_date, tz=None))
        self.all_sessions = self.valid_days

        # numeric-friendly time
        # we can perform binary search on the calendar by introducing this
        self.numerical_valid_days: np.ndarray = convert_dates_to_int(self.valid_days)

        # format-friendly time
        self.market_opens: pd.DatetimeIndex = self.valid_days + pd.Timedelta(hours=self.market_open_time.hour,
                                                                             minutes=self.market_open_time.minute)
        self.market_break_starts: pd.DatetimeIndex = self.valid_days + pd.Timedelta(
            hours=self.market_break_start_time.hour,
            minutes=self.market_break_start_time.minute)
        self.market_break_ends: pd.DatetimeIndex = self.valid_days + pd.Timedelta(hours=self.market_break_end_time.hour,
                                                                                  minutes=self.market_break_end_time.minute)
        self.market_closes: pd.DatetimeIndex = self.valid_days + pd.Timedelta(hours=self.market_close_time.hour,
                                                                              minutes=self.market_close_time.minute)

        # Unix-based time
        self.opens_nanos = self.market_opens.values.astype(np.int64)
        self.break_starts_nanos = self.market_break_starts.values.astype(np.int64)
        self.break_ends_nanos = self.market_break_ends.values.astype(np.int64)
        self.closes_nanos = self.market_closes.values.astype(np.int64)

        # references
        self.schedule = pd.DataFrame(
            index=self.all_sessions,
            data=OrderedDict([
                ("market_open", self.market_opens),
                ("break_start", self.market_break_starts),
                ("break_end", self.market_break_ends),
                ("market_close", self.market_closes)
            ]),
            dtype='datetime64[ns]',
        )

    @property
    def name(self):
        return "上海证券交易所"

    @property
    def sessions(self):
        return self.valid_days

    @property
    def sessions_nanos(self):
        return self.valid_days.values.astype(np.int64)

    @property
    def opens(self):
        return self.market_opens

    @property
    def closes(self):
        return self.market_closes

    @property
    def break_starts(self):
        return self.market_break_starts

    @property
    def break_ends(self) -> pd.DatetimeIndex:
        return self.market_break_ends

    @property
    def first_session(self) -> pd.Timestamp:
        """First calendar session."""
        return self.sessions[0]

    @property
    def last_session(self) -> pd.Timestamp:
        """Last calendar session."""
        return self.sessions[-1]

    @property
    def first_session_open(self) -> pd.Timestamp:
        """Open time of calendar's first session."""
        return self.opens[0]

    @property
    def last_session_close(self) -> pd.Timestamp:
        """Close time of calendar's last session."""
        return self.closes[-1]

    # TODO: clear duplicated elements
    @staticmethod
    def minute_to_session(dt: pd.Timestamp):
        return pd.Timestamp(year=dt.year, month=dt.month, day=dt.day)

    @staticmethod
    def convert_dt_to_session(dt: pd.Timestamp):
        return pd.Timestamp(year=dt.year, month=dt.month, day=dt.day)

    def find(self, dt: typing.Union[pd.Timestamp, pd.DatetimeIndex]) -> typing.Union[typing.List, int]:
        """
        Find the largest `trade_date` that is less or equal than the `dt`.
        """
        if isinstance(dt, pd.DatetimeIndex):
            ndt = convert_dates_to_int(dt)
        else:
            ndt = convert_date_to_int(dt)
        # Here we should minus one since `arr.searchsorted(dt)` returns the largest i, that satisfies
        # arr[i-1] is less or equal than dt, thus the returned i is the value we wanted add 1.
        # To get the needed index, we should minus one.
        indexes = self.numerical_valid_days.searchsorted(ndt, side="right") - 1
        # Note that if dt is less than all the trade_dates, the returned value is -1, which is not well defined,
        # In this case, we should carefully check this value.
        if indexes.ndim == 0:
            return int(indexes)
        else:
            return list(indexes)

    def shift(self, dt: pd.Timestamp, offset=0) -> pd.Timestamp:
        """
        if offset is 0, then we perform the following operations
        if dt is a trade_date:
            return dt
        else:
            return the largest trade_date that is less than dt.
        In all, shift(dt) returns the largest trade_date that is less or equal than dt.
        $$ sup_{t}\{t: t<=dt, t \in trade_dates \}$$.

        After finding the proper date, we perform the shift operation.
        """
        dt = pd.to_datetime(dt)
        dt_revision_index = self.find(dt)
        dt_revision_index += offset
        return self.valid_days[dt_revision_index]

    def history(self, dt: pd.Timestamp, count: int) -> pd.DatetimeIndex:
        """
        First we shift the dt to the true
        """
        assert count != 0
        dt = pd.to_datetime(dt)
        dt_index = self.find(dt)
        return pd.DatetimeIndex(self.valid_days[dt_index-count+1:dt_index+1])

    def sessions_in_range(
            self,
            start_session_label: pd.Timestamp,
            end_session_label: pd.Timestamp,
    ):
        return self.all_sessions[
            self.all_sessions.slice_indexer(
                start_session_label,
                end_session_label
            )
        ]

    def session_open(self, session_label: pd.Timestamp):
        return self.schedule.at[
            session_label,
            "market_open"]

    def session_break_start(self, session_label):
        break_start = self.schedule.at[
            session_label,
            'break_start'
        ]
        return break_start

    def session_break_end(self, session_label):
        break_end = self.schedule.at[
            session_label,
            'break_end'
        ]
        return break_end

    def session_close(self, session_label):
        return self.schedule.at[
            session_label,
            'market_close'
        ]

    def is_session(self):
        # TODO: implement this
        raise NotImplementedError

    @staticmethod
    def _create_monthly_groups(trade_dates):
        rlt = defaultdict(list)
        for key, value in zip(trade_dates, [(x.year, x.month) for x in trade_dates]):
            rlt[value].append(key)
        return rlt

    @staticmethod
    def _create_weekly_groups(trade_dates):
        rlt = defaultdict(list)
        for key, value in zip(trade_dates, [(x.year, x.week) for x in trade_dates]):
            rlt[value].append(key)
        return rlt
    
    @staticmethod
    def Weekly(trade_dates=None, days: typing.Iterable[int]=[-1]):
        temp = XSHGExchangeCalendar._create_weekly_groups(trade_dates)
        rebalance_dates = sorted([x[i] for x in temp.values() for i in days])
        return rebalance_dates
    
    @staticmethod
    def Monthly(trade_dates=None, days: typing.Iterable[int]=[-1]):
        temp = XSHGExchangeCalendar._create_monthly_groups(trade_dates)
        rebalance_dates = sorted([x[i] for x in temp.values() for i in days])
        return rebalance_dates
    
    @staticmethod
    def MonthlyRebalance(data, days=[-1]):
        rebalance_dates = XSHGExchangeCalendar.Monthly(data.index)
        return data.loc[rebalance_dates]
    
    @staticmethod
    def WeeklyRebalance(data, days=[-1]):
        rebalance_dates = XSHGExchangeCalendar.Weekly(data.index)
        return data.loc[rebalance_dates] 

    def RangeWeekly(self, start_date, end_date, days: typing.Iterable[int]=[-1]):
        trade_dates = XSHGExchangeCalendar.sessions_in_range(self, start_date, end_date)
        return XSHGExchangeCalendar.Weekly(trade_dates, days)
     
    def RangeMonthly(self, start_date, end_date, days: typing.Iterable[int]=[-1]):
        trade_dates = XSHGExchangeCalendar.sessions_in_range(self, start_date, end_date)
        return XSHGExchangeCalendar.Monthly(trade_dates, days)
    

class DateIterator(object):
    def __init__(self, start_date, trading_calendar=None):
        if trading_calendar is None:
            self.trading_calendar = XSHGExchangeCalendar(start_date=start_date)
        self.i = self.trading_calendar.find(start_date)
        if self.i < 0:
            self.i = 0

    @property
    def current_session(self):
        return self.trading_calendar.valid_days[self.i]

    def __next__(self):
        self.i += 1

    def __iter__(self):
        return self


DEFAULT_CALENDAR = XSHGExchangeCalendar()


if __name__ == "__main__":
    test = XSHGExchangeCalendar()
    test.find(pd.date_range(start="20050101", periods=10))
    test.history(10, pd.Timestamp("20051110"))
    test.shift(pd.Timestamp("20050104"), 1)
    di = DateIterator(pd.Timestamp("20050101"))
    print(di.current_session)




#!/usr/bin/python3
import argparse
import os

class LatexCli:
    def __init__(self):
        self.parser = argparse.ArgumentParser(description="Alphagens command line tool")
        self.subparsers = self.parser.add_subparsers(dest="command")


        self.subparsers.add_parser("clear", help="clear the outputs")

        # Define the "set" command
        set_parser = self.subparsers.add_parser("set", help="Set something")
        # ... add arguments for set ...

        # Define the "list" command
        list_parser = self.subparsers.add_parser("list", help="List items")
        # ... add arguments for list ...

        # Define the "create" command

        # create_parser = self.subparsers.add_parser("create", help="create the dependent files")

        # create_parser.add_argument("")

        run_parser = ...

        create_parser = self.subparsers.add_parser("create", help="create latex environment")
        
        # Add --type and --name arguments for install command

        create_parser.add_argument("path", 
                            type=str, 
                            help="Specify the path where item should be create")
        
        create_parser.add_argument("--title", 
                                    type=str, 
                                    required=True, 
                                    help="Specify the the title of pdf")
        
        create_parser.add_argument("--name", 
                                    type=str, 
                                    required=True, 
                                    help="Specify the name of pdf")

    def run(self):
        args = self.parser.parse_args()
        # Handle the arguments
        if args.command == "set":
            self.handle_set(args)
        elif args.command == "list":
            self.handle_list(args)
        elif args.command == "install":
            self.handle_install(args)
        elif args.command == "create":
            self.handle_create(args)
        elif args.command == "clear":
            self.handle_clear(args)
        else:
            self.parser.print_help()

    def handle_set(self, args):
        # Your set command implementation
        pass

    def handle_list(self, args):
        # Your list command implementation
        pass

    def handle_install(self, args):
        item_type = args.type
        item_name = args.name
        print(f"Installing {item_type}: {item_name}...")
        # ... Rest of your install command implementation ..

    def handle_create(self, args):
        latex = LatexGenerator(args.path, args.title, args.name)
        latex.deploy()
        print("hello")

    def handle_clear(self, args):
        os.system("rm -rf ./src")
        os.system("rm main.tex")
        os.system("rm main.pdf")
        os.system("rm Makefile")
        os.system("rm -rf ./build")

class LatexGenerator:
    main_include_packages = f"""\\documentclass[12pt]{{article}}

\\usepackage[UTF8]{{ctex}}

\\usepackage{{amsmath}}
\\usepackage{{amsthm}}
\\usepackage{{amsfonts}}
\\usepackage{{amssymb}}
\\usepackage{{geometry}}
\\usepackage{{graphicx}}
\\usepackage{{hyperref}}
\\usepackage{{enumerate}}
\\usepackage{{fancyhdr}} 
\\usepackage{{tcolorbox}}
\\usepackage{{booktabs}} 
\\usepackage{{caption}} 
\\usepackage{{subcaption}}

% cumtom symbols
\\input{{./src/symbols.tex}}

% set page margin
\\setlength{{\\headheight}}{{15pt}}
\\geometry{{
a4paper,
top=2.5cm,
bottom=2.5cm,
left=2cm,
right=2cm
}}

% set header and footer
\\pagestyle{{fancy}}
\\fancyhf{{}}
\\rhead{{nanyi}}
\\lhead{{\\textrm{{homework}}}}
\\rfoot{{\\textrm{{Page}} \\thepage}}


\\begin{{document}}
    """


    main_ends = f"""
\\newpage
% \\bibliographystyle{{siam}}
% \\bibliography{{./src/references}}
\end{{document}}
    """

    symbols = f"""
% src/symbols.tex

% 重定义符号
\\newcommand{{\\R}}{{\mathbb{{R}}}} % 定义\R为实数集符号
"""
    
    Makefile = r"""
LATEX = xelatex
TARGET = main

all = start

start: $(TARGET)

$(TARGET): $(TARGET).tex
	@mkdir -p ./build
	@$(LATEX) -output-directory=./build $@.tex
	@bibtex ./build/$@.aux
	@$(LATEX) -output-directory=./build $@.tex
	@$(LATEX) -output-directory=./build $@.tex
	@ln -sf ./build/$@.pdf $@.pdf

restart:
	@make stop
	@make start

stop:
	@rm -rf ./build

clean:
	@rm -rf ./build
"""

    def __init__(self, path, title, name):
        self.path = os.path.normpath(path)
        self.title = title
        self.name = name

    def generate_infos(self):
        bb = f"""
\\title{{{self.title}}}
\\author{{{self.name}}}
\\date{{\\today}}
\\maketitle
        """
        return bb

    def deploy(self):
        
        print(f"generating {self.path}/main.tex")
        with open(f"{self.path}/main.tex", "w") as file:
            main_text = self.main_include_packages + self.generate_infos() + self.main_ends
            file.write(main_text)
        
        print(f"generating {self.path}/Makefile")
        with open(f"{self.path}/Makefile", "w") as file:
            text = self.Makefile
            file.write(text)
        
        print(f"running mkdir {self.path}/src")
        os.mkdir(f"{self.path}/src")

        print(f"generating {self.path}/src/symbols.tex")
        with open(f"{self.path}/src/symbols.tex", "w") as file:
            text = self.symbols
            file.write(text)
    
        print("successfully")

def main():
    cli = LatexCli()
    cli.run()

if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
import http.client
import traceback
import urllib

import gzip
from io import BytesIO

HTTP_OK = 200
HTTP_AUTHORIZATION_ERROR = 401
class Client:
    domain = 'api.wmcloud.com'
    port = 443
    token = ''
    #设置因网络连接，重连的次数
    reconnectTimes=2
    httpClient = None
    def __init__( self ):
        self.httpClient = http.client.HTTPSConnection(self.domain, self.port, timeout=600)
    def __del__( self ):
        if self.httpClient is not None:
            self.httpClient.close()
    def encodepath(self, path):
        #转换参数的编码
        start=0
        n=len(path)
        re=''
        i=path.find('=',start)
        while i!=-1 :
            re+=path[start:i+1]
            start=i+1
            i=path.find('&',start)
            if(i>=0):
                for j in range(start,i):
                    if(path[j]>'~'):
                        re+=urllib.parse.quote(path[j])
                    else:
                        re+=path[j]  
                re+='&'
                start=i+1
            else:
                for j in range(start,n):
                    if(path[j]>'~'):
                        re+=urllib.parse.quote(path[j])
                    else:
                        re+=path[j]  
                start=n
            i=path.find('=',start)
        return re
    def init(self, token):
        self.token=token
    def getData(self, path):
        result = None
        path='/data/v1' + path
        print (path)
        path=self.encodepath(path)
        for i in range(self.reconnectTimes):
            try:
                #set http header here
                self.httpClient.request('GET', path, headers = {"Authorization": "Bearer " + self.token,
                                                                "Accept-Encoding": "gzip, deflate"})
                #make request
                response = self.httpClient.getresponse()
                result = response.read()
                compressedstream = BytesIO(result)  
                gziper = gzip.GzipFile(fileobj=compressedstream)
                try:
                    result = gziper.read()
                except:
                    pass
                return response.status, result
            except Exception as e:
                if i == self.reconnectTimes-1:
                    raise e
                if self.httpClient is not None:
                    self.httpClient.close()
                self.httpClient = http.client.HTTPSConnection(self.domain, self.port, timeout=60)
        return -1, result


from abc import ABC, abstractmethod
from collections import deque
import numpy as np
from typing import TypeVar

ActType = TypeVar("ActType")
ObsType = TypeVar("ObsType")
BUFFER_SIZE = 50


class BaseAgent(ABC):
    def __init__(self, data, buffer_size=None):
        self.data = data
        if buffer_size:
            self.buffer = deque(maxlen=buffer_size)
        else:
            self.buffer = deque(maxlen=BUFFER_SIZE)

    @abstractmethod
    def take_action(self, state: ObsType):
        raise NotImplementedError("should be implemented in the derived class !")
    

    


from abc import ABCMeta, abstractmethod
from pandas_market_calendars import get_calendar
import itertools
from collections import defaultdict
from typing import Sized, Iterable


class Calendar(object, metaclass=ABCMeta):
    def __init__(self, start_date, end_date):
        self.start_date = start_date
        self.end_date = end_date
        self.exchange = 'SSE'
        self.calendar = get_calendar(self.exchange)
        self.trade_dates = self.calendar.valid_days(start_date=self.start_date, end_date=self.end_date)
        self.trade_dates = self.trade_dates.tz_convert(None)

    def create_monthly_groups(self):
        rlt = defaultdict(list)
        for key, value in zip(self.trade_dates, [(x.year, x.month) for x in self.trade_dates]):
            rlt[value].append(key)
        return rlt

    def create_weekly_groups(self):
        rlt = defaultdict(list)
        for key, value in zip(self.trade_dates, [(x.year, x.week) for x in self.trade_dates]):
            rlt[value].append(key)
        return rlt


class Daily(object):
    def __init__(self, start_date, end_date):
        self.start_date = start_date
        self.end_date = end_date
        self.rebalance_dates = None
        self._create_trade_dates()

    def _create_trade_dates(self):
        calendar = Calendar(self.start_date, self.end_date)
        self.rebalance_dates = calendar.trade_dates

    @property
    def data(self):
        return self.rebalance_dates


class Weekly(object):
    def __init__(self, start_date, end_date, days: Iterable[int]):
        self.start_date = start_date
        self.end_date = end_date
        self.days = days
        self.rebalance_dates = None
        self._create_trade_dates()

    def _create_trade_dates(self):
        calendar = Calendar(self.start_date, self.end_date)
        temp = calendar.create_weekly_groups()
        self.rebalance_dates = sorted([x[i] for x in temp.values() for i in self.days])

    @property
    def data(self):
        return self.rebalance_dates


class Monthly(object):
    def __init__(self, start_date, end_date, days: Iterable[int]):
        self.start_date = start_date
        self.end_date = end_date
        self.days = days
        self.rebalance_dates = None
        self._create_trade_dates()

    def _create_trade_dates(self):
        calendar = Calendar(self.start_date, self.end_date)
        temp = calendar.create_weekly_groups()
        self.rebalance_dates = sorted([x[i] for x in temp.values() for i in self.days])

    @property
    def data(self):
        return self.rebalance_dates



import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
import typing
from ..utils.visualization import ColorGenerator

class StrategyMetrics:

    """用staticmethod好一些吗? 目前数据格式不是很统一, 还是一个一个算比较好
    """
    annulized_factor: int = 252

    @staticmethod
    def annulized_mean(strategy_returns, annulized_factor=252):
        return np.mean(strategy_returns) * annulized_factor
    
    @staticmethod
    def annulized_std(strategy_returns, annulized_factor=252):
        return np.std(strategy_returns) * np.sqrt(annulized_factor)

    @staticmethod
    def annulized_sharpe_ratio(strategy_returns, rfr=0, annulized_factor=252):
        return (np.mean(strategy_returns- rfr)) / np.std(strategy_returns) * np.sqrt(annulized_factor)
    
    @staticmethod
    def drawdown(strategy_returns: pd.Series) -> (pd.Series, float, int):
        """
        Returns
        -------
        (max drawdown, longest drawdown duration, drawdown series,)
        """
        net_curve = (1+strategy_returns).cumpord()
        cmax = net_curve.cumax()
        rlt = cmax / net_curve - 1
        longest_drawdown = cmax.value_counts().max() - 1
        if isinstance(strategy_returns, pd.Series):
            rlt = pd.Series(rlt, index=strategy_returns.index)
        elif isinstance(strategy_returns, np.ndarray):
            rlt = np.array(rlt)
        return rlt.max(), longest_drawdown, rlt
    
    @staticmethod
    def evaluate(strategy_returns, benchmark_returns=0, rfr=0, annulized_factor=252):
        raise NotImplementedError
        result = {}
        result["mean"] = StrategyMetrics.annulized_mean(strategy_returns, annulized_factor)
        result["std"] = StrategyMetrics.annulized_std(strategy_returns, annulized_factor)
        result["sharpe_ratio"] = StrategyMetrics.annulized_sharpe_ratio(strategy_returns, rfr, annulized_factor)

    @staticmethod
    def reg_stats(strategy_returns, benchmark_returns, rfr=0, annulize_factor=252):
        y = strategy_returns - rfr
        X = benchmark_returns.reshape(-1, 1) - rfr
        model = LinearRegression()
        model.fit(X, y)
        annulized_alpha = model.intercept_ * annulize_factor
        beta = model.coef_[0]
        residuals = y - model.predict(X)
        annulized_residual_std = np.std(residuals) * np.sqrt(annulize_factor)
        return annulized_alpha, beta, annulized_residual_std

    @staticmethod
    def plot_heatmap_of_monthly_returns(
        strategy_returns: pd.Series, 
        figsize: tuple[float] = (12, 10),
        title: str = 'Monthly Returns Heatmap',
        vlim: tuple[float] = (-0.3, 0.3),
        dpi: int = 100,
    ):
        tmp = strategy_returns.resample("M").sum()
        tmp.name = "monthly returns"
        tmp = pd.DataFrame(tmp)
        tmp['year'] = tmp.index.year
        tmp['month'] = tmp.index.month
        bb = tmp.pivot_table(tmp, index=tmp["year"], columns=tmp["month"])
        bb.columns.name = None
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize, dpi=dpi)
        ax = sns.heatmap(bb.values, annot=True, cmap="RdYlGn_r", fmt=".2%", linewidths=0.5, linecolor='white', 
                         yticklabels=bb.index, xticklabels=bb.columns, vmin=vlim[0], vmax=vlim[1], ax=ax)
        ax.set_title(title)
        ax.set_xlabel('Month')
        ax.set_ylabel('Year')
        fig.tight_layout()
        return fig, ax

    @staticmethod
    def integrated_informations_by_year(
        strategy_returns,
    ):
        pass

    @staticmethod
    def beta():
        raise Warning("to get beta, call reg_stats")
    
    @staticmethod
    def alpha():
        raise Warning("to get alpha, call reg_stats")
    

class FactorMetrics:
    
    @staticmethod
    def ic_series(factor_scores, forward_returns, method="pearson") -> pd.Series:
        """_summary_
        caculate the 
        - `mean-IC`
        - `ICIR`

        Parameters
        ----------
        factor_scores : _type_
            _description_
        forward_returns : _type_
            _description_
        method : str, optional
            _description_, by default "pearson"

        Returns
        -------
        pd.Series
            _description_
        """
        pass

    @staticmethod
    def plot_pnl_by_groups(strategies_pnls: pd.DataFrame, dpi=100):
        colors = ColorGenerator(5).data
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,8), dpi=dpi)
        for idx, name in enumerate(strategies_pnls.columns):
            pnl = strategies_pnls[name]
            ax.plot(pnl, label=f"group_{idx}", color=colors[idx])
        ax.legend()
        ax.set_xlabel("Date")
        ax.set_ylabel("pnl")
        ax.grid()
        fig.tight_layout()
        return fig, ax

    @staticmethod
    def cum_ic_series():
        raise NotImplementedError

    @staticmethod
    def long_short_returns(integrated_factors) -> pd.Series:
        pass

    @staticmethod
    def group_returns(integrated_factors) -> pd.DataFrame:
        pass

    @staticmethod
    def turnover():
        raise NotImplementedError
    
    @staticmethod
    def fitness():
        pass

    staticmethod
    def margin():
        pass
    

